import unittest

import numpy as np

from osaft.core.backgroundfields import (
    BackgroundField,
    WaveTypes,
    WrongWaveTypeError,
)
from osaft.core.fluids import InviscidFluid, ViscoelasticFluid, ViscousFluid
from osaft.core.functions import exp, pi
from osaft.tests.basetest import BaseTest


class TestBackgroundFieldWrappers(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.pi = pi

        self.inviscid = InviscidFluid(
            self.f, self.rho_f,
            self.c_f,
        )

        self.viscous = ViscousFluid(
            self.f,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
        )

        self.viscoelastic = ViscoelasticFluid(
            self.f,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.eta_p,
            self.zeta_p,
            self.lambda_M,
        )

        self.wave_type = WaveTypes.STANDING
        self.position = pi / 4

        self.cls_inviscid = BackgroundField(
            self.inviscid,
            self.p_0,
            self.wave_type,
            self.position,
        )
        self.cls_viscous = BackgroundField(
            self.viscous,
            self.p_0,
            self.wave_type,
            self.position,
        )
        self.cls_viscoelastic = BackgroundField(
            self.viscoelastic,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.list_cls = [
            self.cls_inviscid, self.cls_viscous,
            self.cls_viscoelastic,
        ]

    def test_f(self) -> None:
        self._f.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.f, self.f)
        self.assertEqual(self.cls_viscous.f, self.f)
        self.assertEqual(self.cls_viscoelastic.f, self.f)

    def test_c_f(self) -> None:
        self._c_f.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.c_f, self.c_f)
        self.assertEqual(self.cls_viscous.c_f, self.c_f)
        self.assertEqual(self.cls_viscoelastic.c_f, self.c_f)

    def test_rho_f(self) -> None:
        self._rho_f.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.rho_f, self.rho_f)
        self.assertEqual(self.cls_viscous.rho_f, self.rho_f)
        self.assertEqual(self.cls_viscoelastic.rho_f, self.rho_f)

    def tests_eta_f(self) -> None:
        self._eta_f.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.eta_f, 0)
        self.assertEqual(self.cls_viscous.eta_f, self.eta_f)
        self.assertEqual(self.cls_viscoelastic.eta_f, self.eta_f)

    def tests_zeta_f(self) -> None:
        self._zeta_f.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.zeta_f, 0)
        self.assertEqual(self.cls_viscous.zeta_f, self.zeta_f)
        self.assertEqual(self.cls_viscoelastic.zeta_f, self.zeta_f)

    def tests_eta_p(self) -> None:
        self._eta_p.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.eta_p, 0)
        self.assertEqual(self.cls_viscous.eta_p, 0)
        self.assertEqual(self.cls_viscoelastic.eta_p, self.eta_p)

    def tests_zeta_p(self) -> None:
        self._zeta_p.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.zeta_p, 0)
        self.assertEqual(self.cls_viscous.zeta_p, 0)
        self.assertEqual(self.cls_viscoelastic.zeta_p, self.zeta_p)

    def tests_lambda_M(self) -> None:
        self._lambda_M.change()
        self.assign_parameters()
        self.assertEqual(self.cls_inviscid.lambda_M, 0)
        self.assertEqual(self.cls_viscous.lambda_M, 0)
        self.assertEqual(self.cls_viscoelastic.lambda_M, self.lambda_M)


class TestViscousBackgroundField(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.fluid_cls = ViscousFluid(
            self.f, self.rho_f,
            self.c_f, self.eta_f,
            self.eta_f,
        )

        self.cls = BackgroundField(
            self.fluid_cls, self.p_0,
            self.wave_type, self.position,
        )

        self.list_cls = [self.fluid_cls, self.cls]
# -----------------------------------------------------------------------------
# Methods
# -----------------------------------------------------------------------------

    def compute_abs_pos(self):
        return self.cls.position / self.cls.k_f.real

    def compute_A(self):
        omega = self.cls.omega
        rho_f = self.rho_f
        p_0 = self.p_0
        return p_0 / (1j * omega * rho_f)

    def compute_abs_A_squared(self):
        A = self.compute_A()
        return np.absolute(A)**2

    def compute_A_in(self, n) -> complex:

        A = self.compute_A()

        wave_type = self.wave_type
        position = self.cls.abs_pos
        k_f = self.cls.k_f

        if wave_type == WaveTypes.TRAVELLING:
            return A * (2 * n + 1) * 1j ** n
        elif wave_type == WaveTypes.STANDING:
            out = exp(1j * position * k_f)
            out += (-1)**n * exp(-1j * position * k_f)
            out *= A / 2 * (2 * n + 1) * 1j**n
            return out

    def compute_E(self):
        return 1 / 4 * (self.p_0**2 * self.fluid_cls.kappa_f)

    def compute_I(self):
        out = abs(self.compute_A() ** 2)
        out *= self.rho_f * self.c_f
        out *= (self.fluid_cls.k_f ** 2) / 2
        return out

# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------

    def test_abs_pos(self):
        self.do_testing(lambda: self.cls.abs_pos, self.compute_abs_pos)

    def test_k_v(self):
        self.do_testing(
            lambda: self.fluid_cls.k_v,
            lambda: self.cls.k_v,
        )

    def test_kappa_f(self):
        self.do_testing(
            lambda: self.fluid_cls.kappa_f,
            lambda: self.cls.kappa_f,
        )

    def test_wave_type(self):
        self.assertRaises(
            ValueError, BackgroundField,
            self.fluid_cls, self.p_0, WaveTypes.STANDING,
        )

    def test_A(self):
        self.do_testing(func_1=self.compute_A, func_2=lambda: self.cls.A)

    def test_abs_A_squared(self):
        self.do_testing(
            func_1=self.compute_abs_A_squared,
            func_2=lambda: self.cls.abs_A_squared,
        )

    def test_A_in(self) -> None:
        for n in range(4, 0, -1):
            self.do_testing(
                func_1=self.compute_A_in, args_1=n,
                func_2=self.cls.A_in, args_2=n,
            )

    def test_A_in_error(self) -> None:
        self.cls.wave_type = 'None'
        self.assertRaises(WrongWaveTypeError, self.cls.A_in, 20)

    def test_E(self) -> None:
        self.do_testing(lambda: self.cls.E_ac, self.compute_E)

    def test_error_E(self) -> None:
        self.cls.wave_type = 'wrong_wave_type'
        with self.assertRaises(WrongWaveTypeError):
            self.cls.E_ac

    def test_I(self) -> None:
        self.wave_type = WaveTypes.TRAVELLING
        self._wave_type.list_of_values = [WaveTypes.TRAVELLING]
        self.do_testing(self.compute_I, lambda: self.cls.I_ac)

    def test_error_I(self) -> None:
        self.cls.wave_type = 'wrong_wave_type'
        with self.assertRaises(WrongWaveTypeError):
            self.cls.I_ac


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from functools import partial
from typing import Optional, Union

import numpy as np

from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import cos, exp, full_range, sin
from osaft.core.helper import InputHandler
from osaft.core.variable import PassiveVariable


class BaseScattering(ABC):
    """ Base class for the scattering field that defines the common
    interface.

    .. note::
       This base class can be used for axisymmetrical models only.

    :param N_max: End of the infinite series for the scattering coefficients
    """

    def __init__(self, N_max: int = 5) -> None:
        """Constructor method
        """

        # Independent variables
        self._N_max = PassiveVariable(N_max, 'number of modes N')

    # -------------------------------------------------------------------------
    # API - Acoustic Velocity and Displacement
    # -------------------------------------------------------------------------

    def radial_acoustic_fluid_velocity(
            self,
            r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            scattered: bool,
            incident: bool,
            mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """ Returns the value for the radial acoustic velocity in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param scattered: scattered field contribution
        :param incident: incident field contribution
        :param mode: specific mode number of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=False,
        )
        radial_func = partial(
            self.V_r,
            scattered=scattered,
            incident=incident,
        )
        out = self.radial_mode_superposition(
            radial_func, r, theta, t, mode,
        )

        return out

    def tangential_acoustic_fluid_velocity(
            self,
            r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            scattered: bool,
            incident: bool,
            mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """ Returns the value for the tangential acoustic velocity in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param scattered: scattered field contribution
        :param incident: incident field contribution
        :param mode: specific mode number of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=False,
        )
        radial_func = partial(
            self.V_theta,
            scattered=scattered,
            incident=incident,
        )
        out = self.tangential_mode_superposition(
            radial_func, r, theta, t, mode,
        )

        return out

    @abstractmethod
    def radial_particle_velocity(
        self,
        r: Union[float, Sequence],
        theta: Union[float, Sequence],
        t: Union[float, Sequence],
        mode: Optional[int] = None,
    ) -> float:
        """ Returns the value for the radial particle velocity in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        pass

    @abstractmethod
    def tangential_particle_velocity(
            self,
            r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            mode: Optional[int] = None,
    ) -> float:
        """ Returns the value for the tangential particle velocity in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        pass

    def radial_particle_displacement(
            self, r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """Particle displacement in radial direction

        Returns the value of the particle displacement
        in radial direction in [m]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` that all
                     modes until :attr:`.N_max`
        """
        velocity = self.radial_particle_velocity(r, theta, t, mode)
        return velocity / (- 1j * self.omega)

    def tangential_particle_displacement(
            self, r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """Particle displacement in tangential direction

        Returns the value of the particle displacement
        in tangential direction in [m]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` that all
                     modes until :attr:`.N_max`
        """
        velocity = self.tangential_particle_velocity(r, theta, t, mode)
        return velocity / (- 1j * self.omega)

    # -------------------------------------------------------------------------
    # Getters and Setters
    # -------------------------------------------------------------------------

    @property
    def N_max(self):
        """Cutoff mode number for infinite sums

        :getter: returns number of infinite sum term
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._N_max.value

    @N_max.setter
    def N_max(self, value):
        self._N_max.value = value

    # -------------------------------------------------------------------------
    # Mode Superposition
    # -------------------------------------------------------------------------

    def radial_mode_superposition(
            self, radial_func: Callable[[int, float], complex],
            r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            mode: int = None,
    ) -> Union[complex, np.ndarray]:
        """Returns either a single mode (``mode=int``) or a the sum until
        :attr:`.N_max` (``mode=None``).

        If ``mode=int`` the formula is

        .. math::
           \\exp(\\text{i}\\omega t)\\,
           f_{\\text{radial, mode}}(r, \\theta)
           \\,P_{\\text{mode}}(\\cos\\theta)

        If ``mode=None`` the formula is

        .. math::
           \\exp(\\text{i}\\omega t)\\sum_{n=0}^{\\text{N}_{\\text{max}}}
           \\,f_{\\text{radial, n}}(r, \\theta)
           \\,P_{\\text{n}}(\\cos\\theta)

        """

        out = exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.cos_monomial(mode, theta, radial_func(mode, r))
        else:
            out *= Leg.cos_poly(
                theta,
                np.array(
                    [radial_func(n, r) for n in full_range(0, self.N_max)],
                ),
            )
        return out

    def tangential_mode_superposition(
            self, radial_func: Callable[[int, float], complex],
            r: Union[float, Sequence],
            theta: Union[float, Sequence],
            t: Union[float, Sequence],
            mode: int,
    ) -> Union[complex, np.ndarray]:
        """Returns either a single mode (``mode=int``) or a the sum until
        :attr:`.N_max` (``mode=None``).

        If ``mode=int`` the formula is

        .. math::
           \\exp(\\text{i}\\omega t)\\,
           f_{\\text{tangential, mode}}(r, \\theta)
           \\,P^{\\prime}_{\\text{mode}}(\\cos\\theta)

        If ``mode=None`` the formula is

        .. math::
           \\exp(\\text{i}\\omega t)\\sum_{n=0}^{\\text{N}_{\\text{max}}}
           \\,f_{\\text{tangential, n}}(r, \\theta)
           \\,P^{\\prime}_{\\text{n}}(\\cos\\theta)

        """

        out = exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.first_cos_monomial(
                mode,
                theta, radial_func(mode, r),
            )
        else:
            out *= Leg.first_cos_poly(
                theta,
                np.array([
                    radial_func(n, r)
                    for n in full_range(0, self.N_max)
                ]),
            )
        return out

    # -------------------------------------------------------------------------
    # Velocity Amplitudes
    # -------------------------------------------------------------------------

    def V_r_i(
        self, n: int,
        r: Union[float, Sequence],
    ) -> complex:
        """Radial incident field velocity term of mode `n`
        without Legendre coefficients

        Returns radial incident field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.A_in(n) * self.k_f * Sp.d1_besselj(n, self.k_f * r)

    def V_theta_i(
        self, n: int,
        r: Union[float, Sequence],
    ) -> complex:
        """Tangential incident field velocity term of mode n
        without Legendre coefficients

        Returns tangential incident field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.A_in(n) * Sp.besselj(n, self.k_f * r) / r

    @abstractmethod
    def V_r_sc(
        self, n: int,
        r: Union[float, Sequence],
    ) -> complex:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        pass

    @abstractmethod
    def V_theta_sc(
        self, n: int,
        r: Union[float, Sequence],
    ) -> complex:
        """Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        pass

    def V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and incident:
            return self.V_r_i(n, r) + self.V_r_sc(n, r)
        elif scattered and not incident:
            return self.V_r_sc(n, r)
        elif not scattered and incident:
            return self.V_r_i(n, r)
        else:
            raise ValueError(
                'Neither scattered nor incident field has '
                'been has been selected. Velocity field is zero.',
            )

    def V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and incident:
            return self.V_theta_i(n, r) + self.V_theta_sc(n, r)
        elif scattered and not incident:
            return self.V_theta_sc(n, r)
        elif not scattered and incident:
            return self.V_theta_i(n, r)
        else:
            raise ValueError(
                'Neither scattered nor incident field has '
                'been has been selected. Velocity field is zero.',
            )


class BaseScatteringRigidParticle(BaseScattering):
    """ Base class for the Scattering Field for a model with a rigid
    particle that defines the common interface.

    This base class is used for axisymmetrical models.

    :param N_max: End of the infinite series for the scattering coefficients
    """

    def __init__(self, N_max):
        super().__init__(N_max)

    @abstractmethod
    def particle_velocity(self, t: float) -> float:
        """Particle velocity

        Returns the value of the particle velocity
        in the direction of the axis of rotational
        symmetry of the radiation field in [m/s]

        :param t: time [s]
        """
        pass

    def radial_particle_velocity(
        self, r: Union[float, Sequence],
        theta: Union[float, Sequence],
        t: Union[float, Sequence],
        mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """Particle velocity in radial direction

        Returns the value of the particle velocity
        in radial direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` that all
                     modes until :attr:`.N_max`
        """
        if mode is None or mode == 1:
            r, theta, t = InputHandler.handle_input(
                r, theta, t,
                self.R_0, inside_sphere=True,
            )
            return self.particle_velocity(t) * cos(theta)
        else:
            return 0 * r * theta * t

    def tangential_particle_velocity(
        self, r: Union[float, Sequence],
        theta: Union[float, Sequence],
        t: Union[float, Sequence],
        mode: Optional[int] = None,
    ) -> Union[complex, np.ndarray]:
        """Particle velocity in tangential direction

        Returns the value of the particle velocity
        in tangential direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific mode number of interest; if `None` that all
                     modes until :attr:`.N_max`
        """

        if mode is None or mode == 1:
            r, theta, t = InputHandler.handle_input(
                r, theta, t,
                self.R_0, inside_sphere=True,
            )
            return -self.particle_velocity(t) * sin(theta)
        else:
            return 0 * r * theta * t


if __name__ == '__main__':
    pass



from abc import ABC
from collections.abc import Callable
from typing import Any, Optional

import numpy as np

from osaft import log


class BaseVariable(ABC):
    """ BaseVariable class defining basic methods and attributes shared between
    active and passive attributes.

    This class and its children are based on the Observer Design Pattern.
    In the present context value changes shall be tracked so
    :class:`ActiveVariable` know when their dependencies changed, and they
    need to recompute their underlying value.

    :param name: Name of the variable
    :raises: TypeError if :attr:`name` is of wrong type
    """

    def __init__(self, name: Optional[str] = None) -> None:
        """Constructor method
        """
        self._value = None
        self._variables_to_notify: list[BaseVariable] = []

        if name is None:
            self._name = 'Unnamed'
        elif isinstance(name, str):
            self._name = name
        else:
            raise TypeError(f'name is of type {type(name)} instead of str')

    @property
    def name(self) -> str:
        """Returns name of the variable.

        :setter: sets :attr:`~osaft.variable.BaseVariable._name`
        """
        return self._name

    @property
    def value(self) -> Any:
        """Value of the attribute

        :setter: just defined in :class:`PassiveVariable`
        :getter: returns/computes the value of the instance
        """
        return self._value

    def notify(self) -> None:
        """Notifies all items in :attr:`BaseVariable._variables_to_notify` that
        a dependent value changed. Just the information of a changed dependency
        is passed.
        """
        log.debug(f'Variable {repr(self)} notifies:')
        for att in self._variables_to_notify:
            log.debug(f'attribute: {att}')
            att._change_update_status()

    def _link(self, input_variable: Any) -> None:
        """Links the inputs to self, so it knows who needs to update when
        :attr:`BaseVariable.value` changes.

        :param input_variable:
        """
        log.info(
            f'Variables linked: {str(input_variable)} depends on'
            f' {str(self)}',
        )
        self._variables_to_notify.append(input_variable)

    @staticmethod
    def _check_types(
        inputs: tuple,
    ) -> None:
        """ Checks if elements of :attr:`inputs` are of type
        :class:`BaseVariable`

        :param inputs: iterable to be checked
        :raises: TypeError if :attr:`inputs` is of wrong type
        """
        for item in inputs:
            if not isinstance(item, BaseVariable):
                out = f'{item} is of type {type(item)} instead of BaseVariable'
                raise TypeError(out)


class PassiveVariable(BaseVariable):
    """ Passive variable without any computation dependencies

    :param val: initial value
    :param name: Name of the variable
    """

    def __init__(
        self, val: Any, name: Optional[str] = None,
    ) -> None:
        """Constructor method
        """
        super().__init__(name)
        self.value = val

        # no if clause needed because last in inheritance
        log.debug(f'Creating {self}')

    @BaseVariable.value.setter
    def value(self, value: Any) -> None:
        """
        Checks if :attr:`BaseVariable._value` is already equal to value.
        If not, first :attr:`BaseVariable._value` is set to `value`
        and then all item in :attr:`BaseVariable._variables_to_notify` are
        notified.

        :param value: new value
        """
        has_changed = True
        if self._value is not None:
            try:
                L1 = len(self._value)
            except TypeError:
                L1 = 0
            try:
                L2 = len(value)
            except TypeError:
                L2 = 0

            if L1 == 0 and L2 == 0:
                has_changed = self._value != value
            elif L1 == L2:
                has_changed = not np.alltrue(self._value == value)

        if has_changed:
            log.debug(f'value change of {repr(self)} to {value}')
            self._value = value
            self.notify()

    def __repr__(self) -> str:
        return f'PassiveVariable(name={self.name}, value={self.value})'

    def __str__(self):
        return f'{self.name} = {self.value}'


class ActiveVariable(BaseVariable):
    """ Variables that depended on at least one :class:`.BaseVariable`

    :param functionCall: function signature that defines how
        :attr:`ActiveVariable._value` is computed
        Callable[None, Any]
    :param name: Name of the variable
    """

    def __init__(
        self, functionCall: Callable, name: Optional[str] = None,
    ) -> None:
        """Constructor method
        """
        super().__init__(name)
        self.__compute = functionCall
        self._needs_update = True

        # no if clause needed because last in inheritance
        log.debug(f'Creating {self}')

    @property
    def needs_update(self) -> bool:
        """ Returns if an update is needed for the next value retrieval
        """
        return self._needs_update

    @property
    def value(self) -> Any:
        """Returns the value of the attribute. If
        :attr:`ActiveVariable.needs_update` is true, it calls
        :meth:`ActiveVariable._compute` first and then changes
        :attr:`ActiveVariable._needs_update` to `False` before returning the
        value.
        """
        if self.needs_update:
            self._value = self._compute()
            self._needs_update = False
            log.debug(f'Recomputed: {repr(self)}')
        return self._value

    def _compute(self) -> Any:
        """Executes the assigned routine and outputs its value."""
        log.debug(f'{repr(self)} recomputes its value')
        value = self.__compute()
        return value

    def _change_update_status(self) -> None:
        """Set :attr:`ActiveVariable._needs_update` to `True` and calls
        :meth:`BaseVariable.notify`.
        """
        log.debug(f'{repr(self)} changes _needs_update to True')
        self._needs_update = True
        self.notify()

    def is_computed_by(self, *variables: Any) -> None:
        """ Adds the notification dependencies

        :param variables: variables that are referenced in
            :meth:`~ActiveVariable._compute`
        :raises: TypeError if no arguments are passed
        """
        if len(variables) == 0:
            out = 'is_computed_by() takes at least one positional argument '
            out += '(0 were given)'
            raise TypeError(out)
        self._check_types(variables)
        for variable in variables:
            variable._link(self)

    def __repr__(self) -> str:
        if not self._needs_update:
            return f'ActiveVariable(name={self.name}, value={self.value})'
        else:
            return f'ActiveVariable(name={self.name}, value=N.A.)'

    def __str__(self):
        if not self._needs_update:
            return f'{self.name} = {self.value}'
        else:
            return f'{self.name} = N.A.'


if __name__ == '__main__':
    pass



import unittest

from osaft import Doinikov1994Rigid, WaveTypes
from osaft.tests.basetest import BaseTest
from osaft.tests.basetest_arf import HelperCompareARF


@unittest.skip('Unstable test skipped')
class TestCompareSmallBoundaryLayer(BaseTest, HelperCompareARF):

    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 1e-2
        self.small_viscous_boundary_layer = True
        self.large_viscous_boundary_layer = False
        self.small_particle_limit = True

        # Frequency
        self._f.low = 0.9e5
        self._f.high = 1.1e5
        self.f = 1e5

        # Radius
        self._R_0.high = 12e-6
        self._R_0.low = 8e-6
        self.R_0 = 10e-6

        # Density
        self._rho_s.low = 4.9e3
        self._rho_s.high = 5e3
        self.rho_s = 5.1e3

        # Viscosity
        self._eta_f.low = 1e-7
        self._eta_f.high = 1e-6
        self.eta_f = 5e-7
        self.zeta_f = 0

        self.cls = Doinikov1994Rigid.ARF(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0, WaveTypes.STANDING,
            self.position,
            small_particle_limit=self.small_viscous_boundary_layer,
            large_viscous_boundary_layer=self.large_viscous_boundary_layer,
            small_viscous_boundary_layer=self.small_particle_limit,
        )

        self.compare_cls = Doinikov1994Rigid.ARF(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0, WaveTypes.STANDING,
            self.position,
            small_particle_limit=self.small_particle_limit,
        )

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == '__main__':
    pass



import unittest
from unittest import TestCase

import numpy as np

from osaft.core.functions import pi
from osaft.plotting.grid import Grid


class TestGrid(TestCase):

    def setUp(self) -> None:

        self.r_min = 1
        self.r_max = 10
        self.theta_min = 0
        self.theta_max = 2 * pi
        self.res_1 = 100
        self.res_2 = 50

        # Grid with the same resolution in both dimensions
        self.grid_1_res = Grid(
            self.r_min, self.r_max, self.theta_min,
            self.theta_max, self.res_1,
        )
        # Grid with the same resolution in both dimensions from a tuple
        self.grid_1_from_seq_res = Grid(
            self.r_min, self.r_max, self.theta_min,
            self.theta_max, [self.res_1],
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res = Grid(
            self.r_min, self.r_max, self.theta_min,
            self.theta_max, (self.res_1, self.res_2),
        )

    def test_dr(self):
        dr = (self.r_max - self.r_min) / self.res_1
        self.assertEqual(self.grid_1_res.dr, dr)
        self.assertEqual(self.grid_1_from_seq_res.dr, dr)
        self.assertEqual(self.grid_2_res.dr, dr)

    def test_dtheta(self):
        dtheta_1 = (self.theta_max - self.theta_min) / self.res_1
        dtheta_2 = (self.theta_max - self.theta_min) / self.res_2
        self.assertEqual(self.grid_1_res.dtheta, dtheta_1)
        self.assertEqual(self.grid_1_from_seq_res.dtheta, dtheta_1)
        self.assertEqual(self.grid_2_res.dtheta, dtheta_2)

    def test_arr_r(self):
        arr_r = np.linspace(self.r_min, self.r_max, self.res_1)
        np.testing.assert_array_equal(self.grid_1_res.arr_r, arr_r)
        np.testing.assert_array_equal(self.grid_1_from_seq_res.arr_r, arr_r)
        np.testing.assert_array_equal(self.grid_2_res.arr_r, arr_r)

    def test_arr_theta(self):
        arr_theta_1 = np.linspace(self.theta_min, self.theta_max, self.res_1)
        arr_theta_2 = np.linspace(self.theta_min, self.theta_max, self.res_2)
        np.testing.assert_array_equal(self.grid_1_res.arr_theta, arr_theta_1)
        np.testing.assert_array_equal(
            self.grid_1_from_seq_res.arr_theta,
            arr_theta_1,
        )
        np.testing.assert_array_equal(self.grid_2_res.arr_theta, arr_theta_2)

    def test_ValuerError(self):

        self.assertRaises(
            ValueError, Grid, r_min=self.r_min, r_max=self.r_max,
            theta_min=self.theta_min, theta_max=self.theta_max,
            res=3 * [self.res_1],
        )


if __name__ == '__main__':
    unittest.main()



from osaft import log
from osaft.core.functions import pi
from osaft.core.variable import ActiveVariable, PassiveVariable


class Frequency:
    """Frequency class

    :param frequency: frequency [Hz]
    """

    # -------------------------------------------------------------------------
    # Magic Methods
    # -------------------------------------------------------------------------

    def __init__(self, frequency: float) -> None:
        """Constructor method
        """

        # Independent variables
        self._f = PassiveVariable(frequency, 'frequency f')

        # Initialize dependent variables
        self._omega = ActiveVariable(
            self._compute_omega,
            'angular frequency omega',
        )

        # Dependencies
        self._omega.is_computed_by(self._f)

        log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return f'Frequency(f={self.f})'

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def f(self) -> float:
        """Excitation frequency [Hz].

        :getter: returns the value for the frequency
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._f.value

    @f.setter
    def f(self, value: float) -> None:
        self._f.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def omega(self) -> float:
        """Returns the angular frequency :math:`\\omega` [rad/s].
        """
        return self._omega.value

    # -------------------------------------------------------------------------
    # Methods Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_omega(self) -> float:
        return self.f * 2 * pi


if __name__ == '__main__':
    pass



import unittest

from osaft import King1934
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import cos, exp, sin
from osaft.solutions.Doinikov1994Rigid import ScatteringField
from osaft.tests.basetest_scattering import (
    HelperCompareScattering,
    HelperScattering,
)
from osaft.tests.Doinikov1994Rigid.test_base import TestBase


class TestScattering(HelperScattering, TestBase):

    def setUp(self) -> None:
        super().setUp()

        self._v_particle_threshold = 1e-5
        self._v_fluid_threshold = 1e-3
        self.N_max = 2

        self.cls = ScatteringField(
            self.f, self.R_0, self.rho_s,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
            self.N_max,
        )

    def V_r_sc(self, n, r) -> complex:
        # grad(phi)
        out = self.cls.k_f * self.alpha_n(n)
        out *= Sp.d1_hankelh1(n, self.cls.k_f * r)
        # curl(Psi)
        out -= n * (n + 1) / r * self.beta_n(n) * Sp.hankelh1(
            n, self.cls.k_v *
            r,
        )

        out *= self.cls.A_in(n)
        return out

    def V_theta_sc(self, n, r) -> complex:
        # curl(Psi)
        arg_v = self.cls.k_v * r
        out = Sp.hankelh1(n, arg_v)
        out += arg_v * Sp.d1_hankelh1(n, arg_v)
        out *= -self.beta_n(n)
        # grad(phi)
        out += self.alpha_n(n) * Sp.hankelh1(n, self.cls.k_f * r)

        out *= self.cls.A_in(n) / r
        return out

    # ------------------

    def radial_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return self.particle_velocity(t) * cos(theta)
        else:
            return 0

    def tangential_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return -self.particle_velocity(t) * sin(theta)
        else:
            return 0

    def particle_velocity(self, t):

        out = Sp.besselj(1, self.x)
        out += self.alpha_n(1) * Sp.hankelh1(1, self.x)
        out += 2 * self.beta_n(1) * Sp.hankelh1(1, self.x_v)

        out *= self.cls.A_in(1) * self.cls.k_f * self.rho_t
        out /= self.x

        out *= exp(-1j * self.cls.omega * t)

        return out

    # ------------------
    # generic tests
    # ------------------

    def test_propertiess(self) -> None:
        properties = ['x', 'x_v', 'mu_1', 'mu_2', 'mu_3', 'mu_4']
        self._test_properties(properties)

    def test_methods_n(self) -> None:
        methods = {}
        methods['alpha_n'] = None
        methods['beta_n'] = None
        methods['xi_n'] = None
        methods['gamma_n'] = None

        self._test_methods_n(methods)

    # ------------------

    def alpha_n(self, n: int) -> complex:
        if n > 1:
            out = n * (n + 1)
            out *= Sp.besselj(n, self.x) * Sp.hankelh1(n, self.x_v)

            out -= self.x * self.gamma_n(n) * Sp.d1_besselj(n, self.x)

            out /= self.xi_n(n)

        elif n > 0:
            out = -2 * (1 - self.rho_t**2) * Sp.besselj(1, self.x)
            out *= Sp.hankelh1(1, self.x_v)

            out -= self.mu_1 * self.mu_3
            out /= self.mu_4

        else:
            out = -Sp.besselj(1, self.x) / Sp.hankelh1(1, self.x)

        return out

    def beta_n(self, n: int) -> complex:
        if n > 1:
            out = Sp.d1_besselj(n, self.x) * Sp.hankelh1(n, self.x)
            out -= Sp.besselj(n, self.x) * Sp.d1_hankelh1(n, self.x)
            out *= -self.x / self.xi_n(n)
        elif n > 0:
            out = self.mu_1 * Sp.hankelh1(1, self.x)
            out -= self.mu_2 * Sp.besselj(1, self.x)
            out *= -(1 - self.rho_t) / self.mu_4
        else:
            out = 0

        return out

    def xi_n(self, n: int) -> complex:
        out = -n * (n + 1)
        out *= Sp.hankelh1(n, self.x)
        out *= Sp.hankelh1(n, self.x_v)

        out += self.x * Sp.d1_hankelh1(n, self.x) * self.gamma_n(n)

        return out

    def gamma_n(self, n: int) -> complex:
        out = self.x_v * Sp.d1_hankelh1(n, self.x_v)
        out += Sp.hankelh1(n, self.x_v)
        return out

    # ------------------

    @property
    def mu_1(self) -> complex:
        out = -self.x * Sp.d1_besselj(1, self.x)
        out += self.rho_t * Sp.besselj(1, self.x)
        return out

    @property
    def mu_2(self) -> complex:
        out = -self.x * Sp.d1_hankelh1(1, self.x)
        out += self.rho_t * Sp.hankelh1(1, self.x)
        return out

    @property
    def mu_3(self) -> complex:
        out = (1 + 2 * self.rho_t)
        out *= Sp.hankelh1(1, self.x_v)

        out += self.x_v * Sp.d1_hankelh1(1, self.x_v)
        return out

    @property
    def mu_4(self) -> complex:
        out = 2 * (1 - self.rho_t**2)
        out *= Sp.hankelh1(1, self.x)
        out *= Sp.hankelh1(1, self.x_v)

        out += self.mu_2 * self.mu_3
        return out

    @property
    def x(self):
        return self.R_0 * self.cls.k_f

    @property
    def x_v(self):
        return self.R_0 * self.cls.k_v


class TestScatteringCompareToKing(HelperCompareScattering, TestBase):

    def setUp(self) -> None:
        super().setUp()

        self._scattering_compare_threshold = 1e-2
        self.N_max = 3
        self.eta_f *= 1e-5
        self.zeta_f *= 1e-5
        self.R_0 = 1e-6

        self.cls = ScatteringField(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
            self.N_max,
        )

        self.compare_cls = King1934.ScatteringField(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
            self.N_max,
        )


if __name__ == '__main__':
    unittest.main()



from osaft import Yosioka1955
from osaft.tests.basetest import BaseTest


class TestScattering(BaseTest):

    def setUp(self) -> None:
        super().setUp()

        self.cls = Yosioka1955.BaseYosioka(
            frequency=self.f,
            radius=self.R_0,
            rho_s=self.rho_s, c_s=self.c_s,
            rho_f=self.rho_f, c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @property
    def lambda_rho(self) -> float:
        return self.rho_s / self.rho_f

    @property
    def sigma(self) -> float:
        return self.c_s / self.c_f

    @property
    def x_f(self) -> float:
        return self.R_0 * self.cls.k_f

    @property
    def x_s(self) -> float:
        return self.R_0 * self.cls.k_s

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self) -> None:
        properties = ['lambda_rho', 'sigma', 'x_s', 'x_f']
        self._test_properties(properties)



import unittest

from osaft import Gorkov1962, King1934, WaveTypes, Yosioka1955
from osaft.core.backgroundfields import WrongWaveTypeError
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import conj, pi, sin
from osaft.tests.basetest import BaseTest
from osaft.tests.basetest_arf import HelperCompareARF


class TestARF(BaseTest):

    def setUp(self) -> None:
        super().setUp()

        self.N_max = 5

        self.cls = Yosioka1955.ARF(
            frequency=self.f,
            radius=self.R_0,
            rho_s=self.rho_s, c_s=self.c_s,
            rho_f=self.rho_f, c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @property
    def F(self):
        sigma = self.cls.sigma
        lamb = self.cls.lambda_rho
        x_s = self.cls.x_s

        if self.wave_type == WaveTypes.STANDING:
            if self.cls.bubble_solution:
                out = sigma * x_s ** 3 * (3 * lamb - x_s**2)
                out /= sigma ** 2 * x_s ** 6 + (3 * lamb - x_s ** 2) ** 2
                return out
            else:
                out = lamb + (2 * (lamb - 1) / 3)
                out /= 1 + 2 * lamb
                out -= 1 / (3 * lamb * sigma ** 2)
                return out
        else:
            if self.cls.bubble_solution:
                pass
            else:
                out = (lamb - (1 + 2 * lamb) / (3 * lamb * sigma ** 2)) ** 2
                out += 2 / 9 * (1 - lamb) ** 2
                out *= 1 / (1 + 2 * lamb) ** 2
                return out

    def small_particle_ARF(self):
        if self.wave_type == WaveTypes.STANDING:
            if self.cls.bubble_solution:
                return self._small_bubble_standing()
            else:
                return self._small_particle_standing()
        else:
            if self.cls.bubble_solution:
                return self._small_bubble_travelling()
            else:
                return self._small_particle_travelling()

    def _small_particle_standing(self):
        k = self.cls.k_f
        E = self.cls.E_ac
        R = self.R_0
        F = self.F
        h = self.cls.position
        return pi * R ** 3 * 4 * k * E * sin(2 * h) * F

    def _small_particle_travelling(self):
        R = self.R_0
        I = self.cls.I_ac
        c = self.c_f
        F = self.F
        x = self.cls.x_f
        return pi * R ** 2 * 4 * x ** 4 * I / c * F

    def _small_bubble_travelling(self):
        I = self.cls.I_ac
        sigma = self.cls.sigma
        lamb = self.cls.lambda_rho
        x_s = self.cls.x_s
        R = self.R_0
        c = self.c_f
        res = (3 * lamb - x_s ** 2)
        out = 4 * pi * R ** 2 * x_s ** 4 * I
        out /= c * (sigma ** 2 * x_s ** 6 + res ** 2)
        return out

    def _small_bubble_standing(self):
        E = self.cls.E_ac
        k = self.cls.k_f
        h = self.position
        F = self.F
        return - 4 * pi / k ** 2 * E * sin(2 * h) * F

    def M_n(self, n):
        B_n = self.cls.B_n(n)
        x_s = self.cls.x_s
        return (-1) ** n / (2 * n + 1) * B_n * Sp.besselj(n, x_s)

    def K_n(self, n):
        B_n = self.cls.B_n(n)
        k_s = self.cls.k_s
        x_s = self.cls.x_s
        return (-1) ** n / (2 * n + 1) * B_n * k_s * Sp.d1_besselj(n, x_s)

    def acoustic_radiation_force(self):
        a = self.R_0
        rho = self.rho_f
        lamb = self.cls.lambda_rho
        k = self.cls.k_f
        s = 0
        for n in range(self.cls.N_max):
            s1 = (
                - 2 * pi * a ** 2 * rho * 2 * (n + 1)
                * (self.K_n(n) * conj(self.K_n(n + 1))).real / 2
            )
            s2 = (
                2 * pi * lamb ** 2 * rho * 2 * n * (n + 1) * (n + 2) *
                (self.M_n(n) * conj(self.M_n(n + 1))).real / 2
            )
            s3 = (
                -2 * pi * a * lamb * rho * 2 * (n + 1) * (n + 2) *
                (self.K_n(n) * conj(self.M_n(n + 1))).real / 2
                + 2 * pi * a * lamb * rho * 2 * n * (n + 1)
                * (self.M_n(n) * conj(self.K_n(n + 1))).real / 2
            )
            s4 = (
                - 2 * pi * k ** 2 * a ** 2 * lamb ** 2 * rho
                * 2 * (n + 1) *
                (self.M_n(n) * conj(self.M_n(n + 1))).real / 2
            )
            s += s1 + s2 + s3 + s4
        if self.cls.wave_type == WaveTypes.STANDING:
            return -s
        return s

    # -------------------------------------------------------------------------
    # Test
    # -------------------------------------------------------------------------

    def test_F(self) -> None:
        try:
            self.do_testing(
                func_1=lambda: self.F,
                func_2=lambda: self.cls.F,
            )
        except ValueError:
            pass

    def test_small_particle_solution(self):
        self.cls.small_particle = True
        self.do_testing(
            self.cls.acoustic_radiation_force,
            self.small_particle_ARF,
        )

    def test_small_bubble_solution(self):
        self.cls.small_particle = True
        self.cls.bubble_solution = True
        self.do_testing(
            self.cls.acoustic_radiation_force,
            self.small_particle_ARF,
        )

    def test_value_error_wrong_wavetype(self):
        self.cls.wave_type = 'wrong_wavetype'

        # Small Particle
        self.cls.small_particle = True
        self.assertRaises(
            WrongWaveTypeError,
            self.cls.acoustic_radiation_force,
        )

        # Small Bubble
        self.cls.bubble_solution = True
        self.assertRaises(
            WrongWaveTypeError,
            self.cls.acoustic_radiation_force,
        )

    def test_large_bubble_error(self):
        self.cls.bubble_solution = True
        self.cls.small_particle = False
        self.assertRaises(
            ValueError,
            self.cls.acoustic_radiation_force,
        )

    def test_value_error_F(self):
        # Small Particle
        self.cls.bubble_solution = True
        self.cls.wave_type = WaveTypes.TRAVELLING
        self.assertRaises(WrongWaveTypeError, lambda: self.cls.F)

    def test_M_n(self):
        for n in range(self.n_runs):
            self.do_testing(self.cls.M_n, self.M_n, n, n)

    def test_K_n(self):
        for n in range(self.n_runs):
            self.do_testing(self.cls.K_n, self.K_n, n, n)

    def test_general_arf(self):
        self.do_testing(
            self.acoustic_radiation_force,
            self.cls.acoustic_radiation_force,
        )

    def test_run_large_particle_warning(self):
        self.cls.small_particle = True
        self.R_0 = 2 * pi * self.cls.fluid.lambda_f
        self.assign_parameters()
        self.cls.acoustic_radiation_force()
        self.cls.bubble_solution = True
        self.assign_parameters()
        self.cls.acoustic_radiation_force()


class TestCompareToKing(BaseTest, HelperCompareARF):

    def setUp(self) -> None:

        BaseTest.setUp(self)

        # Setting speed of sound in particle to be large
        self.c_s = 1e3 * self.c_f
        self._c_s.low = 1e3 * self.c_f

        self.arf_compare_threshold = 1e-4

        self.cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
        )
        self.compare_cls = King1934.ARF(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type,
            self.position,
        )

        self.list_cls = [self.cls, self.compare_cls]


class TestCompareToGorkov(BaseTest, HelperCompareARF):

    def setUp(self) -> None:

        BaseTest.setUp(self)

        self.arf_compare_threshold = 1e-3

        self.cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
        )
        self.compare_cls = Gorkov1962.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
        )

        # Setting size of the particle particle to be small
        self.R_0 = self.cls.fluid.lambda_f * 1e-4
        self._R_0.high = self.cls.fluid.lambda_f * 1e-3

        self.list_cls = [self.cls, self.compare_cls]


class TestCompareToSmallParticle(BaseTest, HelperCompareARF):

    def setUp(self) -> None:

        BaseTest.setUp(self)

        self.arf_compare_threshold = 5e-3

        self.cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
        )
        self.compare_cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position, small_particle=True,
        )

        # Setting size of the particle particle to be small
        self.R_0 = self.cls.fluid.lambda_f * 1e-4
        self._R_0.high = self.cls.fluid.lambda_f * 1e-3

        self.list_cls = [self.cls, self.compare_cls]


class TestCompareSmallBubble(BaseTest, HelperCompareARF):

    def setUp(self) -> None:
        BaseTest.setUp(self)

        self.arf_compare_threshold = 5e-3

        self.cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
        )
        self.compare_cls = Yosioka1955.ARF(
            self.f,
            self.R_0, self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type, self.position,
            small_particle=True, bubble=True,
        )

        # Setting size of the particle particle to be small
        self.R_0 = self.cls.fluid.lambda_f * 1e-4
        self._R_0.high = self.cls.fluid.lambda_f * 1e-3

        # Set density to be small inside to be small
        self.rho_s = self.cls.x_f ** 2 * self.rho_f
        self._rho_s.high = 1.1 * self.rho_s

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Union

from osaft import log
from osaft.core.backgroundfields import WaveTypes, WrongWaveTypeError
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sin
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.King1934.scattering import ScatteringField


class ARF(ScatteringField, BaseARF):
    """ARF class for King (1934)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of incident wave (traveling/standing)
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode
    :param small_particle_limit: compute ARF based on small particle limit
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float,
        rho_f: float, c_f: float,
        p_0: float, wave_type: WaveTypes, position: Optional[float] = None,
        N_max: int = 5,
        small_particle_limit: bool = False,
    ) -> None:
        """Constructor method
        """

        # init of parent class
        ScatteringField.__init__(
            self, frequency, radius, rho_s, rho_f, c_f,
            p_0, wave_type, position, N_max,
        )

        # Passive Variable
        self._small_particle_limit = PassiveVariable(
            small_particle_limit,
            'small particle limit',
        )

        # Acitive variables
        self._f_2 = ActiveVariable(
            self._compute_f_2,
            'dipole scattering coefficient f_2',
        )
        self._Phi = ActiveVariable(
            self._compute_Phi,
            'acoustic contrast factor Phi',
        )

        self._f_2.is_computed_by(self.scatterer._rho_s, self.fluid._rho_f)
        self._Phi.is_computed_by(
            self._f_2, self.field._wave_type,
            self._rho_tilde,
        )

        log.info(str(self))
        log.debug(repr(self))

    def __repr__(self):
        return (
            f'{type(self)}(frequency={self.f}, radius={self.R_0}, '
            f'rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, '
            f'p_0={self.p_0}, position={self.position}, '
            f'small_particle_limit={self.small_particle_limit}, '
            f'wave_type={self.wave_type}, N_max={self.N_max})'
        )

    def __str__(self):
        out = 'King\'s solution to the ARF with following properties: \n'
        out += SF.get_str_text(
            'Small Particle Limit', '',
            self.small_particle_limit, '',
        )
        out += SF.get_str_text('Wave type', '', self.wave_type, '')
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text(
            'Position', 'd',
            self.position, 'rad',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_f',
            self.kappa_f, '1/Pa',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        out += 'Other\n'
        out += SF.get_str_text(
            '#modes', 'N_max',
            self.N_max, None,
        )
        return out

    def acoustic_radiation_force(self) -> float:
        """Computes the ARF in Newton [N].
        """
        if not isinstance(self.wave_type, WaveTypes):
            raise WrongWaveTypeError

        if self.small_particle_limit:
            return self._compute_small_particle_limit_arf()
        elif self.wave_type == WaveTypes.STANDING:
            return self._compute_arf_standing()
        else:
            return self._compute_arf_traveling()

    # -------------------------------------------------------------------------
    # Getters and Setters for Independent Variables
    # -------------------------------------------------------------------------
    @property
    def f_2(self) -> float:
        """Dipole scatting coefficient :math:`f_2` [-]
        """
        return self._f_2.value

    @property
    def Phi(self) -> float:
        r"""Acoustic contrast factor :math:`\Phi` [-]
        """
        return self._Phi.value

    @property
    def small_particle_limit(self) -> bool:
        """ ARF for the small particle limit

        :getter: returns if small particle limit is computed
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._small_particle_limit.value

    @small_particle_limit.setter
    def small_particle_limit(self, value: bool) -> None:
        self._small_particle_limit.value = value

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def _compute_f_2(self) -> float:
        return 2 * (self.rho_tilde - 1) / (2 * self.rho_tilde + 1)

    def _compute_Phi(self):
        if self.wave_type == WaveTypes.STANDING:
            return 1 / 3 + self.f_2 / 2
        else:
            out = (1 + 2 / 9 * (1 - 1 / self.rho_tilde) ** 2)
            out /= (2 + 1 / self.rho_tilde) ** 2
            return out

    def _compute_arf_traveling(self) -> float:
        """Computes acoustic radiation force according to King's paper
        for a plane, progressive wave

        Equation (65) in the paper "On the Acoustic
        Radiation Pressure on Spheres"
        """
        first = 1 / self.HH_n(0)
        second = 2 / self.HH_n(1)
        second *= (self.alpha ** 2 - 3 * (1 - 1 / self.rho_tilde)) ** 2
        second /= self.alpha ** 8

        def higher():
            def term(order):
                out = (order + 1) / self.HH_n(order)
                out *= (self.alpha ** 2 - order * (order + 2)) ** 2
                out /= self.alpha ** (4 * order + 4)
                return out

            term_out = 0.0
            if self.N_max >= 2:
                for n in range(2, self.N_max + 1):
                    term_out += term(n)
            return term_out

        out = 2 * pi * self.rho_f * self.field.abs_A_squared
        out *= (first + second + higher())
        out /= self.alpha ** 2

        return out

    def _compute_arf_standing(self) -> float:
        """Computes acoustic radiation force according to King's
        paper for a plane standing wave

        Equation (74) in the paper "On the Acoustic
        Radiation Pressure on Spheres"
        """
        first = self.FFGG(0) / (self.alpha * self.HH_n(0))
        second = - 2 / self.alpha ** 5 * self.FFGG(1) / self.HH_n(1)
        second *= self.alpha ** 2 - 3 * (1 - 1 / self.rho_tilde)

        def higher():
            def term(order):
                out = (- 1) ** order * (order + 1)
                out /= self.alpha ** (2 * order + 3)
                out *= self.FFGG(order) / self.HH_n(order)
                out *= self.alpha ** 2 - order * (order + 2)
                return out

            term_out = 0.0
            if self.N_max >= 2:
                for n in range(2, self.N_max + 1):
                    term_out += term(n)
            return term_out

        out = pi * self.rho_f * self.field.abs_A_squared
        out *= sin(2 * self.position)
        out *= (first + second + higher())

        return out

    def _compute_small_particle_limit_arf(self) -> float:
        """Computes acoustic radiation force according to King's
        paper for small particles.

        Equation (76) (standing wave) and equation (66) (traveling wave)
        in the paper "On the Acoustic Radiation Pressure on Spheres"
        """
        if 10 * self.R_0 > self.field.lambda_f:
            text = 'Wavelength is less than 10 times greater than specified '
            text += 'Radius. Theory might not be valid anymore.'
            log.info(text)

        if self.wave_type == WaveTypes.STANDING:
            out = pi * self.rho_f * self.field.abs_A_squared
            out *= (self.k_f * self.R_0) ** 3
            out *= self.Phi * sin(2 * self.position)
            return out
        else:
            out = 2 * pi * self.rho_f * self.field.abs_A_squared
            out *= self.Phi * (self.k_f * self.R_0) ** 6
            return out

    # -------------------------------------------------------------------------
    # Coefficients
    # -------------------------------------------------------------------------

    def FFGG(self, n: int) -> complex:
        """Returns the coefficient :math:`F_{n+1}F_{n}+G_{n+1}G_{n}` for `n`.

        :param n: order
        """
        out = self.F_n(n, self.alpha) * self.F_n(n + 1, self.alpha)
        out += self.G_n(n, self.alpha) * self.G_n(n + 1, self.alpha)
        return out

    def FGFG(self, n: int) -> complex:
        """Returns the coefficient :math:`F_{n+1}G_{n}-G_{n+1}F_{n}` for `n`.

        :param n: order
        """
        out = self.F_n(n + 1, self.alpha) * self.G_n(n, self.alpha)
        out -= self.G_n(n + 1, self.alpha) * self.F_n(n, self.alpha)
        return out

    def HH_n(self, n: int) -> float:
        """Returns the coefficient :math:`H_{n}^2(n)H_{n}^2(n+1)` for `n` of
        (56) and (57).

        :param n: order
        """
        # King equations (56) and (57)
        # H_n(n) ** 2 * H_n(n + 1) ** 2
        if n == 1:
            out = (self.alpha ** 2 - 3 * (1 - 1 / self.rho_tilde)) ** 2
            out /= self.alpha ** 10
            out += self.FFGG(n) ** 2
            return out
        else:
            out = (self.alpha ** 2 - n * (n + 2)) ** 2
            out /= self.alpha ** (4 * n + 6)
            out += self.FFGG(n) ** 2
            return out


if __name__ == '__main__':
    pass



from __future__ import annotations

from typing import Optional, Union

from osaft import log
from osaft.core.backgroundfields import WaveTypes, WrongWaveTypeError
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sin, sqrt
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.Doinikov1994Compressible.scattering import ScatteringField


class ARF(ScatteringField, BaseARF, BaseSphereFrequencyComposite):
    """Base class for Doinikov (viscous fluid-viscous sphere; 1994)

    .. note::
       For all implemented solution the small particle limit
       :math:`|x| \\ll 1` is assumed.

       If :attr:`small_viscous_boundary_layer` the ARF is computed for the
       limiting case :math:`|x| \\ll 1 \\ll |x_v|`. If
       :attr:`large_viscous_boundary_layer` is set to `True`
       then the ARF is computed for the limiting case
       :math:`|x| \\ll |x_v| \\ll 1`. Setting both option to `False` will
       lead to error since the solution for an arbitrary boundary layer
       thickness is implemented.

       Both values will be logged in the `info.log` when calling
       :meth:`acoustic_radiation_force`. These conditions will not be
       checked at runtime.


    :param frequency: Frequency [Hz]
    :param radius: Radius of the solid [m]
    :param rho_s: Density of the fluid-like solid [kg/m^3]
    :param c_s: Speed of sound of the solid [m/s]
    :param eta_s: shear viscosity of the solid [Pa s]
    :param zeta_s: bulk viscosity of the solid [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity fluid [Pa s]
    :param zeta_f: bulk viscosity fluid [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position within the standing wave field [m]
    :param small_viscous_boundary_layer: :math:`x_v \\gg 1`
    :param large_viscous_boundary_layer: :math:`x_v \\ll 1`
    :param small_particle_limit: using the limiting cases for the ARF
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float, c_s: float,
        eta_s: float, zeta_s: float,
        rho_f: float, c_f: float,
        eta_f: float, zeta_f: float,
        p_0: float,
        wave_type: WaveTypes,
        position: Optional[float] = None,
        small_viscous_boundary_layer: bool = False,
        large_viscous_boundary_layer: bool = False,
        small_particle_limit: bool = True,
        N_max: int = 5,
    ) -> None:
        """Constructor method
        """

        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, frequency, radius)
        ScatteringField.__init__(
            self, frequency, radius,
            rho_s, c_s, eta_s, zeta_s,
            rho_f, c_f, eta_f, zeta_f,
            p_0, wave_type, position, N_max,
        )

        # independent variables
        self._check_small_particle_limit_value(small_particle_limit)
        self._small_particle_limit = PassiveVariable(
            small_particle_limit,
            'small particle limit',
        )
        self._check_boundary_layer(
            small_viscous_boundary_layer,
            large_viscous_boundary_layer,
        )
        self._small_viscous_boundary_layer = PassiveVariable(
            small_viscous_boundary_layer,
            'small viscous boundary layer limit',
        )
        self._large_viscous_boundary_layer = PassiveVariable(
            large_viscous_boundary_layer,
            'large viscous boundary layer limit',
        )

        # Dependent variables
        self._norm_delta = ActiveVariable(
            self._compute_norm_delta,
            'normalized boundary layer '
            'thickness',
        )
        self._mu = ActiveVariable(
            self._compute_mu,
            '(density * viscosity) ratio',
        )
        self._kappa_t = ActiveVariable(
            self._compute_kappa_t,
            'compressibility ratio',
        )
        self._eta_t = ActiveVariable(
            self._compute_eta_t, 'dynamic viscosity ratio',
        )
        self._zeta_t = ActiveVariable(
            self._compute_zeta_t, 'bulk viscosity ratio',
        )
        self._f1 = ActiveVariable(self._compute_f1, 'f1 ARF factor; B.2.1')
        self._f2 = ActiveVariable(self._compute_f2, 'f2 ARF factor; B.2.1')
        self._f3 = ActiveVariable(self._compute_f3, 'f3 ARF factor; B.2.1')
        self._G = ActiveVariable(self._compute_G, 'G ARF factor; eq. 7.9')
        self._S1 = ActiveVariable(self._compute_S1, 'S1 ARF factor; eq. 7.10')
        self._S2 = ActiveVariable(self._compute_S2, 'S2 ARF factor; eq. 7.11')
        self._S3 = ActiveVariable(self._compute_S3, 'S3 ARF factor; eq. 7.12')
        self._f4 = ActiveVariable(self._compute_f4, 'f4 ARF factor; B.2.1')
        self._g1 = ActiveVariable(self._compute_g1, 'g1 ARF factor; B.2.1')
        self._g2 = ActiveVariable(self._compute_g2, 'g2 ARF factor; B.2.1')
        self._g3 = ActiveVariable(self._compute_g3, 'g3 ARF factor; B.2.1')
        self._g4 = ActiveVariable(self._compute_g4, 'g4 ARF factor; B.2.1')
        self._g5 = ActiveVariable(self._compute_g5, 'g5 ARF factor; B.2.1')
        self._g6 = ActiveVariable(self._compute_g6, 'g6 ARF factor; B.2.1')
        self._g7 = ActiveVariable(self._compute_g7, 'g7 ARF factor; B.2.1')

        # Dependencies
        self._G.is_computed_by(self._S1, self._S2, self._S3)
        self._S1.is_computed_by(
            self._kappa_t, self.fluid._rho_f, self.scatterer._rho_f,
        )
        self._S2.is_computed_by(self._f1, self._f2, self._f4)
        self._S3.is_computed_by(
            self._f1, self._f3, self._f4, self.fluid._rho_f,
            self.scatterer._rho_f, self._kappa_t, self._eta_t,
        )
        self._eta_t.is_computed_by(self.scatterer._eta_f, self.fluid._eta_f)
        self._zeta_t.is_computed_by(self.scatterer._zeta_f, self.fluid._zeta_f)
        self._f1.is_computed_by(self._g1, self._g7)
        self._f2.is_computed_by(self._g1, self._g2, self._g5)
        self._f3.is_computed_by(self._g3, self._g6, self._g7)
        self._f4.is_computed_by(self._g4, self._g6, self._g7)
        self._g1.is_computed_by(self._g5, self._g6)
        self._g2.is_computed_by(self._eta_t, self._zeta_t)
        self._g3.is_computed_by(self._eta_t, self._zeta_t)
        self._g4.is_computed_by(self._eta_t, self._zeta_t)
        self._g5.is_computed_by(self._eta_t)
        self._g6.is_computed_by(self._eta_t)
        self._g7.is_computed_by(self._g6)
        self._norm_delta.is_computed_by(self.fluid._delta, self.sphere._R_0)
        self._kappa_t.is_computed_by(
            self.fluid._rho_f, self.fluid._c_f, self.scatterer._rho_f,
            self.scatterer._c_f,
        )
        self._mu.is_computed_by(
            self.fluid._rho_f, self.fluid._eta_f, self.scatterer._rho_f,
            self.scatterer._eta_f,
        )

        # logging
        log.debug(repr(self))
        log.info(str(self))

    def __repr__(self):
        return (
            f'Donikov1994Rigid.ARF(frequency={self.f}, radius={self.R_0}, '
            f'rho_s={self.rho_s}, c_s={self.c_s}, '
            f'eta_s={self.eta_s}, zeta_s={self.zeta_s}, '
            f'rho_f={self.rho_f}, c_f={self.c_f}, '
            f'eta_f={self.eta_f}, zeta_f={self.zeta_f}, '
            f'p_0={self.p_0}, position={self.position}, {self.wave_type}, '
            f'small_viscous_boundary_layer='
            f'{self.small_viscous_boundary_layer}, '
            f'large_viscous_boundary_layer='
            f'{self.large_viscous_boundary_layer}, '
            f'small_particle_limit={self.small_particle_limit}, '
            f'N_max={self.N_max})'
        )

    def __str__(self):
        out = 'Doinikovs\'s  (1994) model (viscous fluid-rigid sphere) for the'
        out += ' ARF with the following properties: \n'
        out += 'Limit Cases\n'
        out += SF.get_str_text(
            'Small particle limit',
            'small_particle_limit',
            self.small_particle_limit, None,
        )
        out += SF.get_str_text(
            'Small delta',
            'small_viscous_boundary_layer',
            self.small_viscous_boundary_layer, None,
        )
        out += 'Backgroundfield\n'
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text('Wavetype', 'wave_type', self.wave_type, None)
        out += SF.get_str_text(
            'Position', 'd',
            self.position, '',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_f',
            self.kappa_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Shear viscosity', 'eta_f',
            self.eta_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Bulk viscosity', 'zeta_f',
            self.zeta_f, '1/Pa',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_s, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_s',
            self.kappa_s, '1/Pa',
        )
        out += SF.get_str_text(
            'Shear viscosity', 'eta_s',
            self.eta_s, '1/Pa',
        )
        out += SF.get_str_text(
            'Bulk viscosity', 'zeta_s',
            self.zeta_s, '1/Pa',
        )
        return out

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def acoustic_radiation_force(self) -> float:
        """Acoustic radiation fore in [N]

        It logs the current values of :attr:`x` and :attr:`x_v`.

        :raises ValueError: Non supported :attr:`~osaft.solutions.\
Doinikov1994Rigid.baseclass.aseDoinikov1994Rigid.wave_type`
        """
        self._check_boundary_layer(
            self.small_viscous_boundary_layer,
            self.large_viscous_boundary_layer,
        )
        if not isinstance(self.wave_type, WaveTypes):
            raise WrongWaveTypeError
        if self.small_viscous_boundary_layer:
            return self._arf_small_particle_small_delta()
        else:
            return self._arf_small_particle_large_delta()

    # -------------------------------------------------------------------------
    # Acoustic radiation force
    # -------------------------------------------------------------------------

    def _arf_small_particle_large_delta(self) -> float:
        """ Eq. 7.1 for WaveTypes.TRAVELLING; Eq. 7.8 for WaveTypes.STANDING
        """
        log.info(
            f'kappa_t << 1 : {self.kappa_t:.3e} << 1\n'
            f'rho_f << rho_s : {self.rho_s:.3e} << {self.rho_s:.3e}\n',
        )

        if self.wave_type == WaveTypes.TRAVELLING:
            out = 5 - 2 * self.kappa_t - 8 * self.f1

            out *= 1 / 6 * self.norm_delta ** 2
        else:
            out = self.G * self._standing_wave_contribution()

        out *= pi * self.field.abs_A_squared * self.rho_f * self.x_0**3
        return out

    def _arf_small_particle_small_delta(self) -> float:
        """ Eq. 6.1 for WaveTypes.TRAVELLING
        """
        log.info(
            f'kappa_t << 1 : {self.kappa_t:.3e} << 1\n'
            f'mu << 1 : {self.mu:.3e} << 1\n',
        )
        if self.wave_type == WaveTypes.TRAVELLING:
            out = (self.rho_s - self.rho_f) ** 2
            out /= (2 * self.rho_s + self.rho_f)**2 * (1 + self.mu)

            out *= self.x_0**3 * self.norm_delta

            out *= 6 * pi * self.rho_f * self.field.abs_A_squared
        else:
            raise NotImplementedError(
                'This solution is not implemented because Doinikov says '
                'in 6.1.2 of the publication that this solution is '
                'the same as Yosioka already computed.\n If you are '
                'interested in this value us the Yosioka class.',
            )
        return out

    def _standing_wave_contribution(self) -> float:
        return sin(2 * self.k_f.real * self.position)

    # -------------------------------------------------------------------------
    # User input checks
    # -------------------------------------------------------------------------

    @staticmethod
    def _check_boundary_layer(small_delta, large_delta) -> None:
        if small_delta and large_delta:
            raise ValueError(
                'small_viscous_boundary_layer and '
                'large_viscous_boundary_layer can\'t both be True',
            )
        if not small_delta and not large_delta:
            raise NotImplementedError(
                'You are trying to use the solution for an arbitrary '
                'boundary layer thickness. However, so far this solution '
                'has not implemented. Feel free to help us to extend this.',
            )

    @staticmethod
    def _check_small_particle_limit_value(value) -> None:
        if not value:
            raise NotImplementedError(
                'You are trying to use the general solution to the ARF '
                'for this model. However, so far just the limiting cases '
                'are implemented. Feel free to help us to extend this.',
            )

    # -------------------------------------------------------------------------
    # Getters and Setters for dependent variables
    # -------------------------------------------------------------------------

    @property
    def small_particle_limit(self) -> bool:
        """Use limiting case for ARF calculation

        :getter: returns if a small particle limit is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        :raises NotImplementedError: If value is set to False; because just the
            limiting cases for the ARF are implemented
        """
        return self._small_particle_limit.value

    @small_particle_limit.setter
    def small_particle_limit(self, value):
        self._check_small_particle_limit_value(value)
        self._small_particle_limit.value = value

    @property
    def small_viscous_boundary_layer(self) -> bool:
        """Use limiting case of a small viscous boundary layer :math:`\\delta`

        :getter: returns if a small viscous boundary layer case is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._small_viscous_boundary_layer.value

    @small_viscous_boundary_layer.setter
    def small_viscous_boundary_layer(self, value):
        self._small_viscous_boundary_layer.value = value

    @property
    def large_viscous_boundary_layer(self) -> bool:
        """Use limiting case of a large viscous boundary layer :math:`\\delta`

        :getter: returns if a large viscous boundary layer case is used
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        """
        return self._large_viscous_boundary_layer.value

    @large_viscous_boundary_layer.setter
    def large_viscous_boundary_layer(self, value):
        self._large_viscous_boundary_layer.value = value

    # -------------------------------------------------------------------------
    # Dependent variables
    # -------------------------------------------------------------------------

    @property
    def eta_t(self):
        return self._eta_t.value

    def _compute_eta_t(self):
        return self.scatterer.eta_f / self.fluid.eta_f

    @property
    def zeta_t(self):
        return self._zeta_t.value

    def _compute_zeta_t(self):
        return self.scatterer.zeta_f / self.fluid.zeta_f

    @property
    def f1(self):
        return self._f1.value

    def _compute_f1(self):
        return self.g1 / self.g7

    @property
    def f2(self):
        return self._f2.value

    def _compute_f2(self):
        if self.g1 == 0:  # pragma: no cover
            return 0
        return self.g2 * self.g5 / self.g1 / 2

    @property
    def f3(self):
        return self._f3.value

    def _compute_f3(self):
        return self.g3 * self.g6 / self.g7 / 10

    @property
    def f4(self):
        return self._f4.value

    def _compute_f4(self):
        return self.g4 * self.g6 / self.g7 / 2

    @property
    def G(self):
        return self._G.value

    def _compute_G(self):
        return (self.S1 + self.S2 + self.S3) / 3

    @property
    def S1(self):
        return self._S1.value

    def _compute_S1(self):
        return 1 - self.kappa_t + 0.9 * (1 - self.rho_s / self.rho_f)

    @property
    def S2(self):
        return self._S2.value

    def _compute_S2(self):
        out = 2 * self.f2 - 1
        out *= 2 * self.f1

        out -= 4 * self.f4
        return out

    @property
    def S3(self):
        return self._S3.value

    def _compute_S3(self):
        out = 3 - 1 / self.eta_t
        out *= -120 * (self.f3 - self.f4)

        out += 50 * self.f1 + 43 - 10 * self.kappa_t

        out *= self.rho_s - self.rho_f
        out /= 30 * self.rho_f * (3 + 2 / self.eta_t)

        return out

    @property
    def g1(self):
        return self._g1.value

    def _compute_g1(self):
        return self.g5 * self.g6

    @property
    def g2(self):
        return self._g2.value

    def _compute_g2(self):
        out = 209 + 148 * self.eta_t + 48 / self.eta_t
        out *= -1 / 9 / self.zeta_t

        out += 19 + 38 * self.eta_t - 12 / self.eta_t
        return out

    @property
    def g3(self):
        return self._g3.value

    def _compute_g3(self):
        out = 25 + 370 * self.eta_t - 80 / self.eta_t
        out *= 1 / 9 / self.zeta_t

        out -= (12 + 19 * self.eta_t + 4 / self.eta_t)
        return out

    @property
    def g4(self):
        return self._g4.value

    def _compute_g4(self):
        out = 5 + 74 * self.eta_t - 16 / self.eta_t
        out *= 1 / 9 / self.zeta_t

        out -= (3 + 4 / self.eta_t)
        return out

    @property
    def g5(self):
        return self._g5.value

    def _compute_g5(self):
        out = (1 - self.eta_t)
        out *= 19 + 16 / self.eta_t
        return out

    @property
    def g6(self):
        return self._g6.value

    def _compute_g6(self):
        return 89 + 48 / self.eta_t + 38 * self.eta_t

    @property
    def g7(self):
        return self._g7.value

    def _compute_g7(self):
        return self.g6**2

    @property
    def norm_delta(self) -> float:
        """ normalized viscous boundary thickness
        :math:`\\tilde{\\delta}=\\frac{\\delta}{R_0}`
        """
        return self._norm_delta.value

    def _compute_norm_delta(self):
        return self.delta / self.R_0

    @property
    def mu(self) -> float:
        """ (density * viscosity) ratio
        """
        return self._mu.value

    def _compute_mu(self):
        out = self.rho_f * self.eta_f
        out /= self.rho_s * self.eta_s
        return sqrt(out)

    @property
    def kappa_t(self) -> float:
        """ compressibility ratio
        """
        return self._kappa_t.value

    def _compute_kappa_t(self):
        return self.kappa_s / self.kappa_f

    @property
    def x_0(self) -> float:
        """ Real part of :attr:`~.x`
        """
        return self.x.real


if __name__ == '__main__':
    pass



from abc import abstractmethod

from osaft import WaveTypes


class HelperARF:
    """Baseclass for testing the ARF computation
    """
    _threshold = 1e-12

    @abstractmethod
    def compute_arf(self):
        pass

    def test_arf(self) -> None:
        self.do_testing(
            self.compute_arf, self.cls.acoustic_radiation_force,
            threshold=self._threshold,
            zero=1e-30,
        )


class HelperStandingARF(HelperARF):

    def test_arf(self) -> None:
        self.wave_type = WaveTypes.STANDING
        self._wave_type.list_of_values = [self.wave_type]
        HelperARF.test_arf(self)


class HelperTravelingARF(HelperARF):

    def test_arf(self) -> None:

        self.wave_type = WaveTypes.TRAVELLING
        self._wave_type.list_of_values = [self.wave_type]
        HelperARF.test_arf(self)


class HelperCompareARF:
    arf_compare_threshold = 1e-12

    def test_comparison_arf(self):
        self.do_testing(
            self.cls.acoustic_radiation_force,
            self.compare_cls.acoustic_radiation_force,
            threshold=self.arf_compare_threshold,
            zero=1e-30,
        )



import os
import subprocess
import unittest

from matplotlib import pyplot as plt


class BaseTestPlotting(unittest.TestCase):

    def setUp(self) -> None:
        self.create_output_folder('OutputPlotting')

    @property
    def output_folder_path(self) -> str:
        """ Absolute folder path for all figure outputs
        """
        return self._output_folder_path

    @property
    def root_folder(self) -> str:
        """ Folder patch of the calling script
        """
        return subprocess.Popen(
            ['git', 'rev-parse', '--show-toplevel'],
            stdout=subprocess.PIPE,
        ).communicate()[0].rstrip().decode('utf-8')

    def _get_absolute_filename(self, name: str, file_type: str) -> os.path:
        """ Return the absolute path of root_folder + name
        """
        return os.path.join(self.output_folder_path, f'{name}.{file_type}')

    def create_output_folder(self, folder_name: str) -> None:
        """ Creates a folder if non-existent relative to the root_folder
        """
        abs_path = os.path.join(self.root_folder, folder_name)
        self._output_folder_path = abs_path
        if not os.path.isdir(abs_path):
            os.mkdir(abs_path)

    def save_fig(self, fig: plt.Figure, name: str, **kwargs) -> None:
        """Saves the figure

        :param fig: figures instance
        :param name: name of the file the figure is saved to
        :param **kwargs: passed to matplotlib.Figure.savefig
        """
        file_name = self._get_absolute_filename(name, 'pdf')
        fig.savefig(file_name, **kwargs)
        plt.close(fig)


if __name__ == '__main__':
    pass



import unittest

import numpy as np

from osaft.core.functions import pi
from osaft.core.helper import InputHandler as IH
from osaft.core.helper import RadiusTester as RT
from osaft.core.helper import ThetaTester as TT


class TestThetaTester(unittest.TestCase):

    def test_ValueError_float(self):
        self.assertRaises(ValueError, TT.test_theta, 1.01 * pi)
        self.assertRaises(ValueError, TT.test_theta, -0.1)

    def test_ValueError_int(self):
        self.assertRaises(ValueError, TT.test_theta, 4)
        self.assertRaises(ValueError, TT.test_theta, -1)

    def test_ValueError_np_ndarray(self):
        lst = [pi * np.random.rand() for _ in np.arange(10)]
        lst = np.asarray(lst)

        lst[2] = 1.01 * pi
        self.assertRaises(ValueError, TT.test_theta, lst)

        lst[2] = -1.01 * pi
        self.assertRaises(ValueError, TT.test_theta, lst)


class TestRadiusTesterInside(unittest.TestCase):

    def setUp(self):
        self.R_0 = 4 / 3

    def test_ValueError_float(self):
        self.assertRaises(ValueError, RT.is_inside_sphere, 1.01 * pi, self.R_0)
        self.assertRaises(ValueError, RT.is_inside_sphere, -0.1, self.R_0)

    def test_ValueError_int(self):
        self.assertRaises(ValueError, RT.is_inside_sphere, 4, self.R_0)
        self.assertRaises(ValueError, RT.is_inside_sphere, -1, self.R_0)

    def test_ValueError_np_ndarray(self):
        lst = [self.R_0 * np.random.rand() for _ in np.arange(10)]
        lst = np.asarray(lst)

        lst[2] = 1.01 * pi
        self.assertRaises(ValueError, RT.is_inside_sphere, lst, self.R_0)

        lst[2] = -1.01 * pi
        self.assertRaises(ValueError, RT.is_inside_sphere, lst, self.R_0)


class TestRadiusTesterOutside(unittest.TestCase):

    def setUp(self):
        self.R_0 = 4 / 3

    def test_ValueError_float(self):
        self.assertRaises(
            ValueError, RT.is_outside_sphere,
            0.1, self.R_0,
        )
        self.assertRaises(ValueError, RT.is_outside_sphere, -0.1, self.R_0)

    def test_ValueError_int(self):
        self.assertRaises(ValueError, RT.is_outside_sphere, 1, self.R_0)
        self.assertRaises(ValueError, RT.is_outside_sphere, -1, self.R_0)

    def test_ValueError_np_ndarray(self):
        lst = [self.R_0 * (1 + np.random.rand()) for _ in np.arange(10)]
        lst = np.asarray(lst)

        lst[2] = 0.99 * self.R_0
        lst[3] = 1.0 * self.R_0
        self.assertRaises(ValueError, RT.is_outside_sphere, lst, self.R_0)

        lst[2] = -1.01 * pi
        self.assertRaises(ValueError, RT.is_outside_sphere, lst, self.R_0)


class TestInputHandle(unittest.TestCase):

    def setUp(self):
        self.R_0 = 4 / 3

    def test_type(self):
        for var in [1, 1.2, [1, 2], (1, 2)]:
            self.assertTrue(isinstance(IH.handle_r(var, 3, True), np.ndarray))
            self.assertTrue(isinstance(IH.handle_r(var, 1, False), np.ndarray))
            self.assertTrue(
                isinstance(
                    IH.handle_theta(var),
                    np.ndarray,
                ),
            )

    def test_ValueError_float(self):
        self.assertRaises(ValueError, IH.handle_r, 0.1, self.R_0, False)
        self.assertRaises(ValueError, IH.handle_r, -0.1, self.R_0, False)
        self.assertRaises(ValueError, IH.handle_r, 2.1, self.R_0, True)
        self.assertRaises(ValueError, IH.handle_r, -2.1, self.R_0, True)

    def test_ValueError_int(self):
        self.assertRaises(ValueError, IH.handle_r, 1, self.R_0, False)
        self.assertRaises(ValueError, IH.handle_r, -1, self.R_0, False)
        self.assertRaises(ValueError, IH.handle_r, 2, self.R_0, True)
        self.assertRaises(ValueError, IH.handle_r, -2, self.R_0, True)

    def test_ValueError_list(self):
        lst = [self.R_0 * (1 + np.random.rand()) for _ in np.arange(10)]

        lst[2] = 0.99 * self.R_0
        lst[3] = 1.0 * self.R_0
        self.assertRaises(ValueError, IH.handle_r, lst, self.R_0, False)

        lst[2] = -1.01 * self.R_0
        self.assertRaises(ValueError, IH.handle_r, lst, self.R_0, True)

    def test_ValueError_np_ndarray(self):
        lst = [self.R_0 * (1 + np.random.rand()) for _ in np.arange(10)]
        lst = np.asarray(lst)

        lst[2] = 0.99 * self.R_0
        lst[3] = 1.0 * self.R_0
        self.assertRaises(ValueError, IH.handle_r, lst, self.R_0, False)

        lst[2] = -1.01 * pi
        self.assertRaises(ValueError, IH.handle_r, lst, self.R_0, False)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.tests.basetest import NumericTestCase


class TestVariables(NumericTestCase):

    def test_hierarchy(self) -> None:
        P1 = PassiveVariable(-3.0)
        P2 = PassiveVariable(+3.0)

        def compute_A():
            return P1.value + P2.value

        A = ActiveVariable(compute_A)
        A.is_computed_by(P1, P2)

        def compute_B():
            return A.value - P2.value - P1.value

        B = ActiveVariable(compute_B)

        B.is_computed_by(A, P1, P2)

        # initial check
        self.assertEqual(0.0, A.value)
        self.assertEqual(0.0, B.value)
        # change a value and check
        P1.value = 0
        self.assertEqual(0.0, B.value)

    def test_exception_when_linking(self):

        P1 = PassiveVariable(-3.0)
        P2 = PassiveVariable(+3.0)

        def compute_A():
            return P1.value + P2.value

        A = ActiveVariable(compute_A)

        A.is_computed_by(P1, P2)

        self.assertRaises(TypeError, A.is_computed_by, P1, 1)

    def test_no_argument_exception(self):

        P1 = PassiveVariable(-3.0)
        P2 = PassiveVariable(+3.0)

        def compute_A():
            return P1.value + P2.value

        A = ActiveVariable(compute_A)

        self.assertRaises(TypeError, A.is_computed_by)

    def test_variable_name(self):

        P1_name = 'P1'
        P1 = PassiveVariable(1, P1_name)
        self.assertEqual(P1.name, P1_name)

        P2 = PassiveVariable(1)
        self.assertEqual(P2.name, 'Unnamed')

        self.assertRaises(TypeError, PassiveVariable, 1, 1)

    def test_non_single_argument(self) -> None:
        # initial assignment
        P1 = PassiveVariable([1, 2], 'list')
        self.assertEqual(P1.value, [1, 2])
        # change list
        P1.value = [2, 3]
        self.assertEqual(P1.value, [2, 3])
        # longer list
        P1.value = [1, 2, 3]
        self.assertEqual(P1.value, [1, 2, 3])
        # same list
        P1.value = [1, 2, 3]
        self.assertEqual(P1.value, [1, 2, 3])
        # np.array
        val = np.asarray([1, 2, 3])
        P1.value = val
        self.assertTrue(np.alltrue(P1.value == val))
        # change from list to single value
        val = -10.0
        P1.value = val
        self.assertEqual(P1.value, val)
        # change from single value to list
        val = np.asarray([-1, 2.7, 3])
        P1.value = val
        self.assertTrue(np.alltrue(P1.value == val))
        # different order in list
        val = np.asarray([-1, 2.7, 3])
        P1.value = np.asarray([2.7, 3, -1])
        self.assertFalse(np.alltrue(P1.value == val))


class TestActiveVariable(unittest.TestCase):

    def test_needs_update(self):
        self.P1 = PassiveVariable(-1)
        self.P2 = PassiveVariable(1.0)
        self.P3 = PassiveVariable(np.pi)

        self.A = ActiveVariable(lambda: True)
        self.A.is_computed_by(self.P1, self.P2, self.P3)
        self.assertTrue(self.A.needs_update)
        _ = self.A.value
        self.assertFalse(self.A.needs_update)
        for P in [self.P1, self.P2, self.P3]:
            P.value = 42
            self.assertTrue(self.A.needs_update)
            _ = self.A.value
            self.assertFalse(self.A.needs_update)


if __name__ == '__main__':
    unittest.main()



import unittest

from osaft.solutions.Doinikov1994Rigid.base import BaseDoinikov1994Rigid
from osaft.tests.basetest import BaseTest


class TestBase(BaseTest):
    def setUp(self) -> None:
        super().setUp()

        self.cls = BaseDoinikov1994Rigid(
            self.f, self.R_0, self.rho_s,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
            self.p_0, self.wave_type,
            self.position,
        )

    @property
    def rho_t(self) -> float:
        return self.rho_f / self.rho_s

    def test_properties(self) -> None:
        properties = ['rho_t']
        self._test_properties(properties)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from scipy.special import exp1, factorial, hankel1, hankel2, jv, lpmv, yv
from sympy.physics.quantum.cg import CG

from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import SpecialFunctions as SP
from osaft.core.functions import (
    cartesian_2_spherical_coordinates,
    cartesian_2_spherical_vector,
)
from osaft.core.functions import clebsch_gordan_coefficient as cg
from osaft.core.functions import (
    cos,
    exp,
    full_range,
    integrate,
    integrate_osc,
    pi,
    sin,
    spherical_2_cartesian_coordinates,
    spherical_2_cartesian_vector,
    sqrt,
    xexpexp1,
)
from osaft.tests.basetest import NumericTestCase


class TestFunctions(NumericTestCase):

    # -------------------------------------------------------------------------
    # Range
    # -------------------------------------------------------------------------

    def test_full_range(self):
        res = full_range(1)
        self.assertEqual(res, range(0, 2))

        res = full_range(-1, 3)
        self.assertEqual(res, range(-1, 4))

        self.assertRaises(ValueError, full_range, 1, 0, 3)

    # -------------------------------------------------------------------------
    # Coordinate Transforms
    # -------------------------------------------------------------------------

    def test_cartesian_2_spherical(self):

        x = 0
        z = 0
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, 0)
        self.assertEqual(theta, 0)

        x = 1
        z = 1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, pi / 4)

        x = -1
        z = 0
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, 1)
        self.assertEqual(theta, -pi / 2)

        x = -1
        z = 1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, -pi / 4)

        x = -1
        z = -1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, -pi * 3 / 4)

    def test_spherical_2_cartesian(self):

        r = 0
        theta = 0
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(z, 0)

        r = 0
        theta = pi / 2
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(z, 0)

        r = 2
        theta = pi / 4
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 2 * sqrt(2) / 2)
        self.assertAlmostEqual(z, 2 * sqrt(2) / 2)

        r = 3
        theta = pi / 4
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 3 * sqrt(2) / 2)
        self.assertAlmostEqual(z, 3 * sqrt(2) / 2)

        r = 2
        theta = pi / 2
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 2)
        self.assertAlmostEqual(z, 0, threshold=1e-10)

        r = 2
        theta = pi
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0, threshold=1e-10)
        self.assertAlmostEqual(z, -2)

    def test_spherical_2_cartesian_velocities(self):

        theta = 3 / 7 * pi
        v_r = 10
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z), spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

        theta = 0
        v_r = 10
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z), spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

        theta = -pi
        v_r = 0
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z), spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

    def test_cartesian_2_spherical_velocities(self):

        theta = 0
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, 0)
        self.assertAlmostEqual(v_theta, v_x)

        theta = pi / 4
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_x * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, v_x * sqrt(2) / 2)

        theta = pi / 2
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_x)
        self.assertAlmostEqual(v_theta, 0)

        theta = 0
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_z)
        self.assertAlmostEqual(v_theta, 0)

        theta = pi / 4
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_z * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, -v_z * sqrt(2) / 2)

        theta = pi / 2
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, 0)
        self.assertAlmostEqual(v_theta, -v_z)

        theta = 0
        v_x = 10
        v_z = 10
        speed = np.hypot(v_x, v_z)
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, speed * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, speed * sqrt(2) / 2)

    # -------------------------------------------------------------------------
    # Bessel Functions
    # -------------------------------------------------------------------------

    def test_besselj_and_derivatives(self):
        def j(n, z):
            return sqrt(pi / 2 / z) * jv(n + 0.5, z)

        n_end = np.random.randint(1, 100, 1)
        for n in np.random.randint(0, n_end, size=10):
            for _ in np.arange(10):
                z = 10 * np.random.randn(1) + 1j * 10 * np.random.randn(1)
                self.assertEqual(j(n, z), SP.besselj(n, z))
                self._assert_equal_complex(
                    self._df(j, n, z), SP.d1_besselj(n, z),
                )
                self._assert_equal_complex(
                    self._d2f(j, n, z), SP.d2_besselj(n, z),
                )
                self._assert_equal_complex(
                    self._d3f(j, n, z), SP.d3_besselj(n, z),
                )
                self._assert_equal_complex(
                    self._d4f(j, n, z), SP.d4_besselj(n, z),
                )

        n = 0
        self._assert_equal_complex(
            self._df(j, n, z), SP.d1_besselj(n, z),
        )
        self._assert_equal_complex(
            self._d2f(j, n, z), SP.d2_besselj(n, z),
        )
        self._assert_equal_complex(
            self._d3f(j, n, z), SP.d3_besselj(n, z),
        )
        self._assert_equal_complex(
            self._d4f(j, n, z), SP.d4_besselj(n, z),
        )

    def test_bessely_and_derivatives(self):
        def y(n, z):
            return sqrt(pi / 2 / z) * yv(n + 0.5, z)

        n_end = np.random.randint(1, 100, 1)
        for n in np.random.randint(0, n_end, size=10):
            for _ in np.arange(10):
                z = 10 * np.random.randn(1) + 1j * 10 * np.random.randn(1)
                self.assertEqual(y(n, z), SP.bessely(n, z))
                self._assert_equal_complex(
                    self._df(y, n, z), SP.d1_bessely(n, z),
                )
                self._assert_equal_complex(
                    self._d2f(y, n, z), SP.d2_bessely(n, z),
                )
                self._assert_equal_complex(
                    self._d3f(y, n, z), SP.d3_bessely(n, z),
                )
                self._assert_equal_complex(
                    self._d4f(y, n, z), SP.d4_bessely(n, z),
                )

        n = 0
        self._assert_equal_complex(
            self._df(y, n, z), SP.d1_bessely(n, z),
        )
        self._assert_equal_complex(
            self._d2f(y, n, z), SP.d2_bessely(n, z),
        )
        self._assert_equal_complex(
            self._d3f(y, n, z), SP.d3_bessely(n, z),
        )
        self._assert_equal_complex(
            self._d4f(y, n, z), SP.d4_bessely(n, z),
        )

    def test_hankelh1_and_derivatives(self):
        def h1(n, z):
            return sqrt(pi / z / 2) * hankel1(n + 0.5, z)

        n_end = np.random.randint(1, 30, 1)
        for n in np.random.randint(0, n_end, size=10):
            for _ in np.arange(10):
                z = 10 * np.random.randn(1) + 1j * 10 * np.random.randn(1)
                self.assertEqual(h1(n, z), SP.hankelh1(n, z))
                self._assert_equal_complex(
                    self._df(h1, n, z), SP.d1_hankelh1(n, z),
                )
                self._assert_equal_complex(
                    self._d2f(h1, n, z), SP.d2_hankelh1(n, z),
                )
                self._assert_equal_complex(
                    self._d3f(h1, n, z), SP.d3_hankelh1(n, z),
                )
                self._assert_equal_complex(
                    self._d4f(h1, n, z), SP.d4_hankelh1(n, z),
                )
                self._assert_equal_complex(
                    self._d5f(h1, n, z), SP.d5_hankelh1(n, z),
                )

        n = 0
        self._assert_equal_complex(
            self._df(h1, n, z), SP.d1_hankelh1(n, z),
        )
        self._assert_equal_complex(
            self._d2f(h1, n, z), SP.d2_hankelh1(n, z),
        )
        self._assert_equal_complex(
            self._d3f(h1, n, z), SP.d3_hankelh1(n, z),
        )
        self._assert_equal_complex(
            self._d4f(h1, n, z), SP.d4_hankelh1(n, z),
        )
        self._assert_equal_complex(
            self._d5f(h1, n, z), SP.d5_hankelh1(n, z),
        )

    def test_hankelh2_and_derivatives(self):
        def h2(n, z):
            return sqrt(pi / z / 2) * hankel2(n + 0.5, z)

        n_end = np.random.randint(1, 30, 1)
        for n in np.random.randint(0, n_end, size=10):
            for _ in np.arange(10):
                z = 10 * np.random.randn(1) + 1j * 10 * np.random.randn(1)
                self.assertEqual(h2(n, z), SP.hankelh2(n, z))
                self._assert_equal_complex(
                    self._df(h2, n, z), SP.d1_hankelh2(n, z),
                )
                self._assert_equal_complex(
                    self._d2f(h2, n, z), SP.d2_hankelh2(n, z),
                )
                self._assert_equal_complex(
                    self._d3f(h2, n, z), SP.d3_hankelh2(n, z),
                )
                self._assert_equal_complex(
                    self._d4f(h2, n, z), SP.d4_hankelh2(n, z),
                )
                self._assert_equal_complex(
                    self._d5f(h2, n, z), SP.d5_hankelh2(n, z),
                )

        n = 0
        self._assert_equal_complex(
            self._df(h2, n, z), SP.d1_hankelh2(n, z),
        )
        self._assert_equal_complex(
            self._d2f(h2, n, z), SP.d2_hankelh2(n, z),
        )
        self._assert_equal_complex(
            self._d3f(h2, n, z), SP.d3_hankelh2(n, z),
        )
        self._assert_equal_complex(
            self._d4f(h2, n, z), SP.d4_hankelh2(n, z),
        )
        self._assert_equal_complex(
            self._d5f(h2, n, z), SP.d5_hankelh2(n, z),
        )

    def test_adaptive_derivative(self):
        def j(n, z):
            return sqrt(pi / 2 / z) * jv(n + 0.5, z)

        def y(n, z):
            return sqrt(pi / 2 / z) * yv(n + 0.5, z)

        def h1(n, z):
            return sqrt(pi / z / 2) * hankel1(n + 0.5, z)

        def h2(n, z):
            return sqrt(pi / z / 2) * hankel2(n + 0.5, z)

        dfdzs = [self._df, self._d2f, self._d3f, self._d4f, self._d5f]

        n_end = np.random.randint(1, 100, 1)
        for n in np.random.randint(0, n_end, size=10):
            for _ in np.arange(10):
                z = 10 * np.random.randn(1) + 1j * 10 * np.random.randn(1)

                for i, df in enumerate(dfdzs, start=1):
                    self._assert_equal_complex(
                        df(j, n, z), SP.adaptive_derivative_besselj(n, z, i),
                    )
                    self._assert_equal_complex(
                        df(y, n, z), SP.adaptive_derivative_bessely(n, z, i),
                    )
                    self._assert_equal_complex(
                        df(h1, n, z), SP.adaptive_derivative_hankelh1(n, z, i),
                    )
                    self._assert_equal_complex(
                        df(h2, n, z), SP.adaptive_derivative_hankelh2(n, z, i),
                    )

        n = 0
        for i, df in enumerate(dfdzs, start=1):
            self._assert_equal_complex(
                df(j, n, z), SP.adaptive_derivative_besselj(n, z, i),
            )
            self._assert_equal_complex(
                df(y, n, z), SP.adaptive_derivative_bessely(n, z, i),
            )
            self._assert_equal_complex(
                df(h1, n, z), SP.adaptive_derivative_hankelh1(n, z, i),
            )
            self._assert_equal_complex(
                df(h2, n, z), SP.adaptive_derivative_hankelh2(n, z, i),
            )

    # -------------------------------------------------------------------------
    # Integration
    # -------------------------------------------------------------------------

    def test_integrate(self):
        def func(x):
            return sin(x)**2

        self.assertAlmostEqual(integrate(func, 0, 2 * pi), pi)
        self.assertAlmostEqual(integrate(func, 0, 2 * pi, 1e-12), pi)

        def func(x):
            return -2 * x + 1

        self.assertAlmostEqual(integrate(func, 0, 2), -2)
        self.assertAlmostEqual(integrate(func, -2, 1 / 2), 6.25)
        self.assertAlmostEqual(integrate(func, -2, 1 / 2, 1e-12), 6.25)

    def test_integrate_osc(self):

        def h1_0(z):
            return SP.hankelh1(0, k_v * z).real

        def h1_1(z):
            return SP.hankelh1(1, k_v * z).real

        def h1_2(z):
            return SP.hankelh1(2, k_v * z).real

        def h1_3(z):
            return SP.hankelh1(3, k_v * z).real

        def j1_0(z):
            return SP.besselj(0, k_v * z).real

        def j1_1(z):
            return SP.besselj(1, k_v * z).real

        def j1_2(z):
            return SP.besselj(2, k_v * z).real

        def j1_3(z):
            return SP.besselj(3, k_v * z).real

        # Parameters
        R0 = 8e-6
        lambda_v = 30e-6
        delta = 50e-6
        x1 = R0
        x2 = R0 + 20 * delta
        k_v = 2 * pi / lambda_v + 1j / delta

        # Zeroth Order Bessel and Hankel
        self.assertAlmostEqual(
            integrate_osc(h1_0, x1, x2, delta, lambda_v)[0],
            8.37973712155099e-7, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_1, x1, x2, delta, lambda_v)[0],
            2.397545981955677e-6, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_2, x1, x2, delta, lambda_v)[0],
            2.219617723321184e-6, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_3, x1, x2, delta, lambda_v)[0],
            4.026448321970975e-7, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_0, x1, x2, delta, lambda_v)[0],
            4.38223531947702, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_1, x1, x2, delta, lambda_v)[0],
            4.659233107792744, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_2, x1, x2, delta, lambda_v)[0],
            -4.310991826110144, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_3, x1, x2, delta, lambda_v)[0],
            -4.750479141883805, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_0, x1, x2, delta)[0],
            8.37973712155099e-7, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_1, x1, x2, delta)[0],
            2.397545981955677e-6, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_2, x1, x2, delta)[0],
            2.219617723321184e-6, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_3, x1, x2, delta)[0],
            4.026448321970975e-7, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_0, x1, x2, delta)[0],
            4.38223531947702, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_1, x1, x2, delta)[0],
            4.659233107792744, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_2, x1, x2, delta)[0],
            -4.310991826110144, threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_3, x1, x2, delta)[0],
            -4.750479141883805, threshold=1e-7,
        )

    # -------------------------------------------------------------------------
    # Clebsch Gordan
    # -------------------------------------------------------------------------

    def test_clebsch_gordan(self):
        def random_value():
            return np.random.randint(0, 4)
        nbr_nonzero = 0

        # Test function for int values

        while nbr_nonzero < 20:
            j1 = random_value()
            j2 = random_value()
            m1 = random_value()
            m2 = random_value()
            j = random_value()
            m = random_value()

            result = cg(j1, m1, j2, m2, j, m)

            self.assertAlmostEqual(
                result,
                float(CG(j1, m1, j2, m2, j, m).doit()),
            )

            if result != 0:
                nbr_nonzero += 1

        # Test errors for non int inputs
        numbers = [1.5, 1, 1, 1, 1, 1]

        for i in np.arange(6):
            j = numbers[i % 6]
            j1 = numbers[(i + 1) % 6]
            j2 = numbers[(i + 2) % 6]
            m = numbers[(i + 3) % 6]
            m1 = numbers[(i + 4) % 6]
            m2 = numbers[(i + 5) % 6]

        self.assertRaises(ValueError, cg, j1, m1, j2, m2, j, m)

    # -------------------------------------------------------------------------
    # Legendre Polynomials
    # -------------------------------------------------------------------------

    def test_legendre(self):

        # first_legendre_cos_poly
        def tmp(theta, coefficients):
            out = 0
            for n, c in enumerate(coefficients):
                out += c * lpmv(1, n, cos(theta))
            return out

        for _ in np.arange(10):
            theta = 2 * pi * np.random.rand()
            coefficients = np.random.rand(10) + 1j * np.random.rand(10)
            n = np.random.randint(0, 5)

            # Test Monomial
            # legendre_cos_monomial
            self.assertEqual(
                Leg.cos_monomial(n, theta, coefficients[0]),
                coefficients[0] * lpmv(0, n, cos(theta)),
            )

            # first_legendre_cos_monomial
            self.assertEqual(
                Leg.first_cos_monomial(n, theta, coefficients[0]),
                coefficients[0] * lpmv(1, n, cos(theta)),
            )

            # Test Polynomials
            theta = 2 * pi * np.random.rand(5)
            for index, th in enumerate(theta):
                # legendre_cos_poly
                self.assertEqual(
                    Leg.cos_poly(theta, coefficients)[index],
                    np.polynomial.legendre.legval(
                        cos(th),
                        coefficients,
                    ),
                )
                # legendre_first_cos_poly
                self.assertAlmostEqual(
                    Leg.first_cos_poly(theta, coefficients)[index],
                    tmp(th, coefficients),
                )

    # -------------------------------------------------------------------------
    # Exponential Integral
    # -------------------------------------------------------------------------

    def test_xexpexp1(self):

        def compare_function(z: complex, N: int = 20):
            if z.real > 20:
                out = 0
                for n in full_range(N):
                    out += factorial(n) / (-z) ** n
                return out
            else:
                return z * exp(z) * exp1(z)

        for x in np.linspace(1, 100, 10):
            z1 = x * (1 + 1j)
            z2 = -1j * z1
            z3 = z1 + z2
            self.assertAlmostEqual(xexpexp1(z1), compare_function(z1))
            self.assertAlmostEqual(xexpexp1(z2), compare_function(z2))
            self.assertAlmostEqual(xexpexp1(z3), compare_function(z3))

    # -------------------------------------------------------------------------
    # Helper
    # -------------------------------------------------------------------------

    @staticmethod
    def _df(func, n, z):
        out = -func(n + 1, z) + (n / z) * func(n, z)
        return out

    @staticmethod
    def _d2f(func, n, z):
        out = -TestFunctions._df(func, n + 1, z)
        out -= n / z**2 * func(n, z)
        out += n / z * TestFunctions._df(func, n, z)
        return out

    @staticmethod
    def _d3f(func, n, z):
        out = -TestFunctions._d2f(func, n + 1, z)
        out += 2 * n / z**3 * func(n, z)
        out -= n / z**2 * TestFunctions._df(func, n, z)
        out -= n / z**2 * TestFunctions._df(func, n, z)
        out += n / z * TestFunctions._d2f(func, n, z)
        return out

    @staticmethod
    def _d4f(func, n, z):
        out = -TestFunctions._d3f(func, n + 1, z)
        out -= 6 * n / z**4 * func(n, z)
        out += 3 * 2 * n / z**3 * TestFunctions._df(func, n, z)
        out -= 3 * n / z**2 * TestFunctions._d2f(func, n, z)
        out += n / z * TestFunctions._d3f(func, n, z)
        return out

    @staticmethod
    def _d5f(func, n, z):
        out = -TestFunctions._d4f(func, n + 1, z)
        out += 24 * n / z**5 * func(n, z)
        out -= 24 * n / z**4 * TestFunctions._df(func, n, z)
        out += 12 * n / z**3 * TestFunctions._d2f(func, n, z)
        out -= 4 * n / z**2 * TestFunctions._d3f(func, n, z)
        out += n / z * TestFunctions._d4f(func, n, z)
        return out

    def _assert_equal_complex(self, z1, z2):
        threshold = 1e-8 * np.abs(z2[0])

        self.assertTrue(np.abs(np.real(z1[0]) - np.real(z2[0])) < threshold)
        self.assertTrue(np.abs(np.imag(z1[0]) - np.imag(z2[0])) < threshold)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from osaft import BackgroundField, InviscidFluid, WaveTypes
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import full_range
from osaft.solutions.base_scattering import (
    BaseScattering,
    BaseScatteringRigidParticle,
)


class Scattering(BaseScattering):

    def __init__(self, N_max: int):
        super().__init__(N_max)
        self.R_0 = 1e-6
        self.k_f = 2e4
        f = 1e6
        self.omega = 2 * np.pi * f
        self.fluid = InviscidFluid(f, 1000, 1500)
        self.field = BackgroundField(self.fluid, 1e5, WaveTypes.TRAVELLING)

    def V_r_sc(self, n, r) -> complex:
        pass

    def V_theta_sc(self, n, r) -> complex:
        pass

    def radial_particle_velocity(self, r, theta, t, mode) -> complex:
        pass

    def tangential_particle_velocity(self, r, theta, t, mode) -> complex:
        pass

    def A_in(self, n):
        return self.field.A_in(n)


class ScatteringRigidParticle(BaseScatteringRigidParticle):
    def __init__(self, N_max: int):
        super().__init__(N_max)
        self.R_0 = 1e-6
        self.k_f = 2e4
        f = 1e6
        self.omega = 2 * np.pi * f
        self.fluid = InviscidFluid(f, 1000, 1500)
        self.field = BackgroundField(self.fluid, 1e5, WaveTypes.TRAVELLING)

    def V_r_sc(self, n, r) -> complex:
        pass

    def V_theta_sc(self, n, r) -> complex:
        pass

    def radial_particle_velocity(self, r, theta, t, mode):
        if mode is None or mode == 1:
            return self.particle_velocity(t) * np.cos(theta)
        else:
            return 0 * r * theta * t

    def tangential_particle_velocity(self, r, theta, t, mode):
        if mode is None or mode == 1:
            return -self.particle_velocity(t) * np.sin(theta)
        else:
            return 0 * r * theta * t

    def particle_velocity(self, t):
        return t


class TestBaseScattering(unittest.TestCase):

    def setUp(self) -> None:
        self.cls = Scattering(5)

    def test_tangential_acoustic_fluid_velocity(self) -> None:
        self.assertRaises(
            ValueError,
            self.cls.tangential_acoustic_fluid_velocity,
            0, 0, 0, False, False,
        )

    def test_radial_acoustic_fluid_velocity(self) -> None:
        self.assertRaises(
            ValueError,
            self.cls.radial_acoustic_fluid_velocity,
            0, 0, 0, False, False,
        )

    def test_N_max(self) -> None:
        self.assertEqual(5, self.cls.N_max)
        self.cls.N_max = 7
        self.assertEqual(7, self.cls.N_max)

    def test_radial_mode_superposition(self) -> None:
        def radial_mode_superposition(r_func, r, theta, t, mode):
            out = np.exp(-1j * self.cls.omega * t)
            if mode is not None:
                out *= Leg.cos_monomial(mode, theta, r_func(mode, r))
            else:
                array = [r_func(n, r) for n in full_range(0, self.cls.N_max)]
                out *= Leg.cos_poly(theta, np.array(array))
            return out

        def radial_func(n, r):
            return Sp.besselj(n, r) - 1j * Sp.hankelh1(n, r)

        for mode in [0, 1, 2, 3, 4, None]:
            r = self.cls.R_0 * (np.random.random() + 1)
            theta = np.pi * np.random.random()
            t = np.random.random()
            with self.subTest(r=r, theta=theta, t=t, mode=mode):
                val1 = radial_mode_superposition(
                    radial_func, r, theta, t, mode,
                )
                val2 = self.cls.radial_mode_superposition(
                    radial_func, r, theta, t, mode,
                )
                self.assertEqual(val1, val2)

    def test_tangential_mode_superposition(self) -> None:
        def tangential_mode_superposition(t_func, r, theta, t, mode):
            out = np.exp(-1j * self.cls.omega * t)
            if mode is not None:
                out *= Leg.first_cos_monomial(mode, theta, t_func(mode, r))
            else:
                array = [t_func(n, r) for n in full_range(0, self.cls.N_max)]
                out *= Leg.first_cos_poly(theta, np.array(array))
            return out

        def tangential_func(n, r):
            return Sp.besselj(n, r) - 1j * Sp.hankelh1(n, r)

        for mode in [0, 1, 2, 3, 4, None]:
            r = self.cls.R_0 * (np.random.random() + 1)
            theta = np.pi * np.random.random()
            t = np.random.random()
            with self.subTest(r=r, theta=theta, t=t, mode=mode):
                val1 = tangential_mode_superposition(
                    tangential_func, r, theta, t, mode,
                )
                val2 = self.cls.tangential_mode_superposition(
                    tangential_func, r, theta, t, mode,
                )
                self.assertEqual(val1, val2)

    def test_V_r_i(self) -> None:
        def V_r_i(n, r):
            out = self.cls.field.A_in(n) * self.cls.k_f
            out *= Sp.d1_besselj(n, self.cls.k_f * r)
            return out

        for n in np.arange(5):
            r = self.cls.R_0 * (np.random.random() + 1)
            with self.subTest(n=n, r=r):
                self.assertEqual(V_r_i(n, r), self.cls.V_r_i(n, r))

    def test_V_theta_i(self) -> None:
        def V_theta_i(n, r):
            out = self.cls.field.A_in(n)
            out *= Sp.besselj(n, self.cls.k_f * r) / r
            return out

        for n in np.arange(5):
            r = self.cls.R_0 * (np.random.random() + 1)
            with self.subTest(n=n, r=r):
                self.assertAlmostEqual(
                    V_theta_i(n, r), self.cls.V_theta_i(n, r),
                )


class TestBaseScatteringRigidParticle(unittest.TestCase):

    def setUp(self):
        self.cls = ScatteringRigidParticle(5)

    def test_radial_particle_velocity(self):
        def radial_particle_velocity(r, theta, t, mode):
            if mode is None or mode == 1:
                return np.cos(theta) * t
            else:
                return 0

        for mode in [0, 1, 2, 3, 4, None]:
            r = np.random.random()
            theta = np.random.random() * np.pi
            t = np.random.random()

            with self.subTest(r=r, theta=theta, t=t, mode=mode):
                self.assertEqual(
                    radial_particle_velocity(r, theta, t, mode),
                    self.cls.radial_particle_velocity(r, theta, t, mode),
                )

    def test_tangential_particle_velocity(self):
        def tangential_particle_velocity(r, theta, t, mode):
            if mode is None or mode == 1:
                return -np.sin(theta) * t
            else:
                return 0

        for mode in [0, 1, 2, 3, 4, None]:
            r = np.random.random()
            theta = np.random.random() * np.pi
            t = np.random.random()

            with self.subTest(r=r, theta=theta, t=t, mode=mode):
                self.assertEqual(
                    tangential_particle_velocity(r, theta, t, mode),
                    self.cls.tangential_particle_velocity(r, theta, t, mode),
                )


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod


class BaseStreaming(ABC):
    """ Base class for the Streaming Field that defines the common
    interface.

    This base class is used for axisymmetrical models.
    """

    @abstractmethod
    def radial_Euler_streaming(
        self, r: float, theta: float, t: float,
    ) -> float:
        """ Returns the value for the radial Euler streaming velocity in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """
        pass

    @abstractmethod
    def tangential_Euler_streaming(
        self, r: float, theta: float, t: float,
    ) -> float:
        """ Returns the value for the tangential Euler streaming velocity in
        [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """
        pass

    @abstractmethod
    def radial_Lagrange_streaming(
        self, r: float, theta: float, t: float,
    ) -> float:
        """ Returns the value for the radial Lagrange streaming velocity in
        [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """
        pass

    @abstractmethod
    def tangential_Lagrange_streaming(
        self, r: float, theta: float,
        t: float,
    ) -> float:
        """ Returns the value for the tangential Lagrange streaming velocity
        in [m/s].

        This method must be implemented by every theory to have a common
        interface for other modules.

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """
        pass


if __name__ == '__main__':
    pass



import matplotlib.pyplot as plt
import numpy as np

from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.arf.arf_plotter import ARFPlotter
from osaft.plotting.datacontainers.arf_datacontainer import ARFData
from osaft.solutions.base_arf import BaseARF


class ARFPlot:
    """Plotting different ARF solutions inside same axis over attribute

    :param attr_name: name of attribute for x-axis
    :param x_values: x values for axis
    """

    def __init__(self, attr_name: str = None, x_values: np.ndarray = None):
        """Constructor method"""
        self.plotter = ARFPlotter()
        self._solutions = {}

        self._attr_name = PassiveVariable(attr_name, 'Name of attribute')
        self._x_values = PassiveVariable(x_values, 'Name of attribute')
        # the callable is not important; we want to use the dependency logic of
        # the ActiveVariable and we are not actually interested in the value
        self._needs_computation = ActiveVariable(
            lambda: True,
            'ARF needs to be recomputed',
        )

        self._needs_computation.is_computed_by(self._attr_name, self._x_values)

    @property
    def attr_name(self) -> str:
        """Attribute that is used as x-axis

        :getter: returns the attribute name for the x-axis
        :setter: sets attribute name
        """
        return self._attr_name.value

    @attr_name.setter
    def attr_name(self, value: str) -> None:
        self._attr_name.value = value

    @property
    def x_values(self) -> np.ndarray:
        """Values of x-axis

        :getter: returns x-axis values
        :setter: sets x-axis values
        """
        return self._x_values.value

    @x_values.setter
    def x_values(self, values: np.ndarray) -> None:
        self._x_values.value = values

    def set_abscissa(self, x_values: np.ndarray, attr_name: str) -> None:
        self.x_values = x_values
        self.attr_name = attr_name

    def add_solution(self, solution: BaseARF) -> None:
        """Add solution to list of solutions for plotting

        :param solution: specific ARF solution, e.g. King1932
        """
        if solution.name in self._solutions:
            raise ValueError(
                'Solution with the same name ({solution.name}) is already ',
                'existent. Consider renaming the attribute `name` of this ',
                'solution({solution}).',
            )
        self._solutions[solution.name] = ARFData(solution)

    def remove_solution(self, solution: BaseARF) -> None:
        """Remove solution of list of solutions for plotting

        :param solution: specific ARF solution, e.g. King1932
        """
        self._solutions.pop(solution.name, None)

    def _compute_arf(self) -> None:
        if not self._needs_computation.needs_update:
            return None

        # value of the ActiveVariable needs to be accessed
        # such that the needs_update property changes to False
        _ = self._needs_computation.value
        for _, items in self._solutions.items():
            items.compute_arf(self.attr_name, self.x_values)

    def _normalize_arf(self, normalization_name: str) -> None:
        if normalization_name is None:
            return None

        norm = self._solutions[normalization_name].arf
        for _, items in self._solutions.items():
            items.normalize_arf(norm)

    def plot_solutions(
            self,
            file_name: str = 'ARF',
            ax: plt.Axes = None,
            normalization_name: str = None,
            plot_method=plt.plot,
            **kwargs,
    ) -> (plt.Figure, plt.Axes):
        """Plot all solutions in stack over attribute set via
        :meth:`set_abscissa()`

        :param file_name: name of output files(s)
        :param ax: axes object where plot will be generated
        :param normalization_name: name of the solutions used for normalizing
        :param plot_method: matplotlib native plotting method (e.g.plt.loglog)
        :param kwargs: keyword arguments that get piped to :attr:`plot_method`
        """

        self._compute_arf()
        self._normalize_arf(normalization_name)

        for _, data in self._solutions.items():
            fig, ax = self.plotter.plot_solution(
                self.x_values, data, file_name, ax, plot_method, **kwargs,
            )

        self.plotter.set_labels(ax, self.attr_name, normalization_name)
        self.plotter.add_legend(ax)

        return fig, ax


if __name__ == '__main__':
    pass



class BaseSolution:
    """ Base class for all solutions with the name property
    """

    @property
    def name(self) -> str:
        """Name of the model used as default for plotting label

        :setter: set the name for the model
        :getter: returns the name of the model
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name
